# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import base64
import os
import sys
import time

from parameterized import parameterized

import mock

from twisted.application import service
from twisted.internet import defer
from twisted.internet import reactor
from twisted.trial import unittest

from buildbot_worker import base
from buildbot_worker import pb
from buildbot_worker import util
from buildbot_worker.test.fake.runprocess import Expect
from buildbot_worker.test.util import command

if sys.version_info >= (3, 6):
    import msgpack
    # pylint: disable=ungrouped-imports
    from buildbot_worker.msgpack import decode_http_authorization_header
    from buildbot_worker.msgpack import encode_http_authorization_header
    from buildbot_worker.msgpack import BuildbotWebSocketClientProtocol
    from buildbot_worker.pb import BotMsgpack  # pylint: disable=ungrouped-imports


class TestHttpAuthorization(unittest.TestCase):
    maxDiff = None
    if sys.version_info < (3, 6):
        skip = "Not python 3.6 or newer"

    def test_encode(self):
        result = encode_http_authorization_header(b'name', b'pass')
        self.assertEqual(result, 'Basic bmFtZTpwYXNz')

        result = encode_http_authorization_header(b'name2', b'pass2')
        self.assertEqual(result, 'Basic bmFtZTI6cGFzczI=')

    def test_encode_username_contains_colon(self):
        with self.assertRaises(ValueError):
            encode_http_authorization_header(b'na:me', b'pass')

    def test_decode(self):
        result = decode_http_authorization_header(
            encode_http_authorization_header(b'name', b'pass'))
        self.assertEqual(result, ('name', 'pass'))

        # password can contain a colon
        result = decode_http_authorization_header(
            encode_http_authorization_header(b'name', b'pa:ss'))
        self.assertEqual(result, ('name', 'pa:ss'))

    def test_contains_no__basic(self):
        with self.assertRaises(ValueError):
            decode_http_authorization_header('Test bmFtZTpwYXNzOjI=')

        with self.assertRaises(ValueError):
            decode_http_authorization_header('TestTest bmFtZTpwYXNzOjI=')

    def test_contains_forbidden_character(self):
        with self.assertRaises(ValueError):
            decode_http_authorization_header('Basic test%test')

    def test_credentials_do_not_contain_colon(self):
        value = 'Basic ' + base64.b64encode(b'TestTestTest').decode()
        with self.assertRaises(ValueError):
            decode_http_authorization_header(value)


class TestException(Exception):
    pass


class FakeStep(object):

    "A fake master-side BuildStep that records its activities."

    def __init__(self):
        self.finished_d = defer.Deferred()
        self.actions = []

    def wait_for_finish(self):
        return self.finished_d

    def remote_update(self, updates):
        for update in updates:
            if 'elapsed' in update[0]:
                update[0]['elapsed'] = 1
        self.actions.append(["update", updates])

    def remote_complete(self, f):
        self.actions.append(["complete", f])
        self.finished_d.callback(None)


class FakeBot(base.BotBase):
    WorkerForBuilder = pb.WorkerForBuilderPbLike


class TestBuildbotWebSocketClientProtocol(command.CommandTestMixin, unittest.TestCase):
    maxDiff = None
    if sys.version_info < (3, 6):
        skip = "Not python 3"

    def setUp(self):
        self.protocol = BuildbotWebSocketClientProtocol()
        self.protocol.sendMessage = mock.Mock()
        self.protocol.factory = mock.Mock()
        self.protocol.factory.password = b'test_password'
        self.protocol.factory.name = b'test_username'

        self.protocol.factory.buildbot_bot.builders = {'test_builder': mock.Mock()}
        self.protocol.dict_def = {}
        self.protocol.sendClose = mock.Mock()

        def mock_util_now(_reactor=None):
            return 0

        # patch util.now function to never let tests access the time module of the code
        self.patch(util, 'now', mock_util_now)

        self.list_send_message_args = []

        def send_message_test(payload, isBinary):
            msg = msgpack.unpackb(payload, raw=False)
            self.list_send_message_args.append(msg)

        self.protocol.sendMessage = send_message_test

    def assert_sent_messages(self, msgs_expected):
        # checks, what messages has been called in sendMessage
        self.assertEqual(msgs_expected, self.list_send_message_args)
        self.list_send_message_args[:] = []

    def setup_with_worker_for_builder(self):
        self.protocol.onOpen()
        # we are not interested in list_send_message_args before onMessage was called by test
        self.list_send_message_args[:] = []
        self.protocol.factory.buildbot_bot = BotMsgpack('test/dir')
        service.MultiService.startService(self.protocol.factory.buildbot_bot)

        self.protocol.factory.buildbot_bot.builder_protocol_command = {'test_builder': None}
        self.protocol.factory.buildbot_bot.builder_basedirs = {'test_builder': 'basedir'}

    @defer.inlineCallbacks
    def test_call_get_worker_info_success(self):
        self.protocol.factory.buildbot_bot.remote_getWorkerInfo = mock.Mock()
        self.protocol.factory.buildbot_bot.remote_getWorkerInfo.return_value = {
            'test': 'data_about_worker'}

        msg = {'op': 'get_worker_info', 'seq_number': 0}
        self.protocol.onMessage(msgpack.packb(msg), True)
        yield self.protocol._deferwaiter.wait()

        self.protocol.factory.buildbot_bot.remote_getWorkerInfo.assert_called()

        msgs_expected = {'op': 'response', 'seq_number': 0, 'result': {'test': 'data_about_worker'}}
        self.assertEqual(self.list_send_message_args, [msgs_expected])

    @defer.inlineCallbacks
    def send_message(self, message):
        self.protocol.onMessage(msgpack.packb(message), True)
        yield self.protocol._deferwaiter.wait()

    @parameterized.expand([
        ('print_op', {'seq_number': 1, 'message': 'test'}),
        ('print_seq_number', {'op': 'print', 'message': 'test'}),
        ('keepalive_op', {'seq_number': 1}),
        ('keepalive_seq_number', {'op': 'keepalive'}),
        ('get_worker_info_op', {'seq_number': 1}),
        ('get_worker_info_seq_number', {'op': 'get_worker_info'}),
        ('start_command_op', {'seq_number': 1}),
        ('start_command_seq_number', {'op': 'start_command'}),
        ('shutdown_op', {'seq_number': 1}),
        ('shutdown_seq_number', {'op': 'shutdown'}),
        ('interrupt_command_op', {'seq_number': 1}),
        ('interrupt_command_seq_number', {'op': 'interrupt_command'}),
        ('response_op', {'seq_number': 1}),
        ('response_seq_number', {'op': 'response'}),
    ])
    @defer.inlineCallbacks
    def test_msg(self, name, msg):
        # if msg does not have 'sep_number' or 'op', response sendMessage should not be called
        with mock.patch('twisted.python.log.msg') as mock_log:
            yield self.send_message(msg)
            mock_log.assert_any_call('Invalid message from master: {}'.format(msg))

        self.assert_sent_messages([])

    @parameterized.expand([
        (
            'start_command', {
                'op': 'start_command',
                'seq_number': 1,
                'command_name': 'test_command',
                'args': 'args'
            },
            'command_id'
        ), (
            'start_command', {
                'op': 'start_command',
                'seq_number': 1,
                'command_id': '123',
                'command_name': 'test_command',
            },
            'args'
        ), (
            'start_command', {
                'op': 'start_command',
                'seq_number': 1,
                'command_id': '123',
                'args': 'args'
            },
            'command_name'
        ), (
            'interrupt_command', {
                'op': 'interrupt_command',
                'seq_number': 1,
                'why': 'test_why'
            },
            'command_id'
        ), (
            'call_print', {
                'op': 'print',
                'seq_number': 1
            },
            'message'
        ), (
            'call_interrupt_command', {
                'op': 'interrupt_command',
                'seq_number': 1,
                'command_id': '123'
            },
            'why'
        ), (
            'call_interrupt_command', {
                'op': 'interrupt_command',
                'seq_number': 1,
                'why': 'test_reason'
            },
            'command_id'
        )])
    @defer.inlineCallbacks
    def test_missing_parameter(self, command, msg, missing_parameter):
        self.protocol.onOpen()
        # we are not interested in list_send_message_args before onMessage was called by test
        self.list_send_message_args[:] = []
        yield self.send_message(msg)
        self.assert_sent_messages([{
            'op': 'response',
            'seq_number': 1,
            'result': '\'message did not contain obligatory "{0}" key\''.format(missing_parameter),
            'is_exception': True
        }])

    @defer.inlineCallbacks
    def test_on_message_unrecognized_command(self):
        self.protocol.onOpen()
        # we are not interested in list_send_message_args before onMessage was called by test
        self.list_send_message_args[:] = []

        yield self.send_message({'op': 'test', 'seq_number': 0})

        self.assert_sent_messages([{
            'is_exception': True,
            'op': 'response',
            'result': 'Command test does not exist.',
            'seq_number': 0
        }])

    def test_authorization_header(self):
        result = self.protocol.onConnecting('test')

        self.assertEqual(result.headers, {
            "Authorization": encode_http_authorization_header(b'test_username', b'test_password')
        })

    @defer.inlineCallbacks
    def test_call_print_success(self):
        self.protocol.factory.buildbot_bot = BotMsgpack('test/dir')
        with mock.patch('twisted.python.log.msg') as mock_log:
            yield self.send_message({'op': 'print', 'seq_number': 0, 'message': 'test_message'})
            mock_log.assert_any_call("message from master:", 'test_message')

        self.assert_sent_messages([{'op': 'response', 'seq_number': 0, 'result': None}])

    @defer.inlineCallbacks
    def test_call_keepalive(self):
        with mock.patch('twisted.python.log.msg') as mock_log:
            yield self.send_message({'op': 'keepalive', 'seq_number': 0})
            mock_log.assert_any_call("Connection keepalive confirmed.")

        self.assert_sent_messages([{'op': 'response', 'seq_number': 0, 'result': None}])

    @defer.inlineCallbacks
    def test_call_start_command_success(self):
        self.setup_with_worker_for_builder()

        # check if directory was created
        with mock.patch('os.makedirs') as mkdir:
            yield self.send_message({
                'op': 'start_command',
                'seq_number': 0,
                'command_id': '123',
                'command_name': 'mkdir',
                'args': {'paths': ['basedir/test_dir'], 'test1': 'value1', 'test2': 'value2'}
            })
            mkdir.assert_called()

    @defer.inlineCallbacks
    def test_call_start_command_failed(self):
        self.patch(time, 'time', lambda: 123.0)
        self.setup_with_worker_for_builder()

        path = os.path.join('basedir', 'test_dir')
        # check if directory was created
        with mock.patch('os.makedirs') as mkdir:
            mkdir.side_effect = OSError(1, 'test_error')
            yield self.send_message({
                'op': 'start_command',
                'seq_number': 1,
                'command_id': '123',
                'command_name': 'mkdir',
                'args': {'paths': [path], 'test1': 'value1', 'test2': 'value2'}
            })
            mkdir.assert_called()

        self.assert_sent_messages([
            {
                'op': 'update',
                'args': [
                    ['rc', 1],
                    ['elapsed', 0],
                    ['header', ['mkdir: test_error: {}\n'.format(path), [35], [123.0]]]
                ],
                'command_id': '123',
                'seq_number': 0
            }, {
                'op': 'complete',
                'args': None,
                'command_id': '123',
                'seq_number': 1
            },
            # response result is always None, even if the command failed
            {'op': 'response', 'result': None, 'seq_number': 1}
        ])

        def create_msg(seq_number):
            return {
                'op': 'response',
                'seq_number': seq_number,
                'result': None
            }

        yield self.send_message(create_msg(0))
        yield self.send_message(create_msg(1))

        # worker should not send any new messages in response to masters 'response'
        self.assertEqual(self.list_send_message_args, [])

    @defer.inlineCallbacks
    def test_call_start_command_shell_success(self):
        self.patch(time, 'time', lambda: 123.0)
        self.setup_with_worker_for_builder()

        # patch runprocess to handle the 'echo', below
        workdir = os.path.join('basedir', 'test_basedir')
        self.patch_runprocess(
            Expect(['echo'], workdir)
            .update('header', 'headers')  # note that this is partial line
            .update('stdout', 'hello\n')
            .update('rc', 0)
            .exit(0)
            )

        yield self.send_message({
            'op': 'start_command',
            'seq_number': 1,
            'command_id': '123',
            'command_name': 'shell',
            'args': {'command': ['echo'], 'workdir': workdir}
        })

        self.assert_sent_messages([
            {
                'op': 'update',
                'args': [
                    ['stdout', ['hello\n', [5], [123.0]]],
                    ['rc', 0],
                    ['elapsed', 0],
                    ['header', ['headers\n', [7], [123.0]]]
                ],
                'command_id': '123',
                'seq_number': 0
            }, {
                'op': 'complete',
                'args': None,
                'command_id': '123',
                'seq_number': 1
            }, {
                'op': 'response', 'seq_number': 1, 'result': None
            }
        ])

    @defer.inlineCallbacks
    def test_call_start_command_shell_success_logs(self):
        self.patch(time, 'time', lambda: 123.0)
        self.setup_with_worker_for_builder()

        workdir = os.path.join('basedir', 'test_basedir')
        self.patch_runprocess(
            Expect(['echo'], workdir)
            .update('header', 'headers\n')
            .update('log', ('test_log', ('hello')))
            .update('log', ('test_log', ('hello1\n')))
            .update('log', ('test_log2', ('hello2\n')))
            .update('log', ('test_log3', ('hello3')))
            .update('rc', 0)
            .exit(0)
            )

        yield self.send_message({
            'op': 'start_command',
            'seq_number': 1,
            'command_id': '123',
            'command_name': 'shell',
            'args': {'command': ['echo'], 'workdir': workdir}
        })

        self.assert_sent_messages([
            {
                'op': 'update',
                'args': [
                    ['header', ['headers\n', [7], [123.0]]],
                    ['log', ['test_log', ['hellohello1\n', [11], [123.0]]]],
                    ['log', ['test_log2', ['hello2\n', [6], [123.0]]]],
                    ['rc', 0],
                    ['elapsed', 0],
                    ['log', ['test_log3', ['hello3\n', [6], [123.0]]]],
                ],
                'command_id': '123',
                'seq_number': 0
            }, {
                'op': 'complete',
                'args': None,
                'command_id': '123',
                'seq_number': 1
            }, {
                'op': 'response', 'seq_number': 1, 'result': None
            }
        ])

    @defer.inlineCallbacks
    def test_start_command_shell_success_updates_single(self):
        self.patch(time, 'time', lambda: 123.0)
        self.setup_with_worker_for_builder()

        # patch runprocess to handle the 'echo', below
        workdir = os.path.join('basedir', 'test_basedir')
        self.patch_runprocess(
            Expect(['echo'], workdir)
            .updates([('header', 'headers'), ('stdout', 'hello\n'), ('rc', 0)])
            .exit(0)
            )

        yield self.send_message({
            'op': 'start_command',
            'seq_number': 1,
            'command_id': '123',
            'command_name': 'shell',
            'args': {'command': ['echo'], 'workdir': workdir}
        })

        self.assert_sent_messages([
            {
                'op': 'update',
                'args': [
                    ['stdout', ['hello\n', [5], [123.0]]],
                    ['rc', 0],
                    ['elapsed', 0],
                    ['header', ['headers\n', [7], [123.0]]]
                ],
                'command_id': '123',
                'seq_number': 0
            }, {
                'op': 'complete',
                'args': None,
                'command_id': '123',
                'seq_number': 1
            }, {
                'op': 'response', 'seq_number': 1, 'result': None
            }
        ])

    @defer.inlineCallbacks
    def test_call_shutdown_success(self):
        # shutdown stops reactor, we can not test it so we just mock
        self.protocol.factory.buildbot_bot.remote_shutdown = mock.Mock()

        yield self.send_message({'op': 'shutdown', 'seq_number': 0})

        self.protocol.factory.buildbot_bot.remote_shutdown.assert_called()

    @defer.inlineCallbacks
    def test_call_interrupt_command_no_command_to_interrupt(self):
        self.setup_with_worker_for_builder()
        self.protocol.factory.command.doInterrupt = mock.Mock()

        with mock.patch('twisted.python.log.msg') as mock_log:
            yield self.send_message({
                'op': 'interrupt_command',
                'seq_number': 1,
                'command_id': '123',
                'why': 'test_reason'
            })
            mock_log.assert_any_call(
                'asked to interrupt current command: {0}'.format('test_reason'))
            mock_log.assert_any_call(' .. but none was running')

        self.protocol.factory.command.doInterrupt.assert_not_called()

    @defer.inlineCallbacks
    def test_call_interrupt_command_success(self):
        self.patch(time, 'time', lambda: 123.0)
        self.setup_with_worker_for_builder()
        self.protocol.factory.command.doInterrupt = mock.Mock()

        # patch runprocess to pretend to sleep (it will really just hang forever,
        # except that we interrupt it)
        workdir = os.path.join('basedir', 'test_basedir')
        self.patch_runprocess(
            Expect(['sleep', '10'], workdir)
            .update('header', 'headers')
            .update('wait', True)
        )

        yield self.send_message({
            'op': 'start_command',
            'seq_number': 1,
            'command_id': '123',
            'command_name': 'shell',
            'args': {'command': ['sleep', '10'], 'workdir': workdir}
        })

        # wait a jiffy..
        d = defer.Deferred()
        reactor.callLater(0.01, d.callback, None)
        yield d

        self.assert_sent_messages([
            {
                'op': 'response',
                'seq_number': 1,
                'result': None
            }
        ])

        yield self.send_message({
            'op': 'interrupt_command',
            'seq_number': 1,
            'command_id': '123',
            'why': 'test_reason'
        })

        self.assert_sent_messages([
            {
                'op': 'update',
                'seq_number': 0,
                'command_id': '123',
                'args': [['header', ['headers\n', [7], [123.0]]]]
            }, {
                'op': 'update',
                'seq_number': 1,
                'command_id': '123',
                'args': [['rc', -1], ['header', ['killing\n', [7], [123.0]]]],
            }, {
                'op': 'complete', 'seq_number': 2, 'command_id': '123', 'args': None
            }, {
                'op': 'response', 'seq_number': 1, 'result': None
            }
    ])



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.internet import defer
from twisted.internet import reactor
from twisted.python import log
from twisted.trial import unittest

from buildbot.test.util.warnings import assertProducesWarnings
from buildbot.util import lineboundaries
from buildbot.warnings import DeprecatedApiWarning


class LBF(unittest.TestCase):

    def setUp(self):
        self.callbacks = []
        with assertProducesWarnings(DeprecatedApiWarning,
                                    message_pattern="does not accept callback anymore"):
            self.lbf = lineboundaries.LineBoundaryFinder(self._callback)

    def _callback(self, wholeLines):
        self.assertEqual(wholeLines[-1], '\n', f'got {repr(wholeLines)}')
        self.callbacks.append(wholeLines)
        d = defer.Deferred()
        reactor.callLater(0, d.callback, None)
        return d

    def assertCallbacks(self, callbacks):
        self.assertEqual(self.callbacks, callbacks)
        self.callbacks = []

    # tests

    @defer.inlineCallbacks
    def test_already_terminated(self):
        yield self.lbf.append('abcd\ndefg\n')
        self.assertCallbacks(['abcd\ndefg\n'])
        yield self.lbf.append('xyz\n')
        self.assertCallbacks(['xyz\n'])
        yield self.lbf.flush()
        self.assertCallbacks([])

    @defer.inlineCallbacks
    def test_partial_line(self):
        for c in "hello\nworld":
            yield self.lbf.append(c)
        self.assertCallbacks(['hello\n'])
        yield self.lbf.flush()
        self.assertCallbacks(['world\n'])

    @defer.inlineCallbacks
    def test_empty_appends(self):
        yield self.lbf.append('hello ')
        yield self.lbf.append('')
        yield self.lbf.append('world\n')
        yield self.lbf.append('')
        self.assertCallbacks(['hello world\n'])

    @defer.inlineCallbacks
    def test_embedded_newlines(self):
        yield self.lbf.append('hello, ')
        self.assertCallbacks([])
        yield self.lbf.append('cruel\nworld')
        self.assertCallbacks(['hello, cruel\n'])
        yield self.lbf.flush()
        self.assertCallbacks(['world\n'])

    @defer.inlineCallbacks
    def test_windows_newlines_folded(self):
        r"Windows' \r\n is treated as and converted to a newline"
        yield self.lbf.append('hello, ')
        self.assertCallbacks([])
        yield self.lbf.append('cruel\r\n\r\nworld')
        self.assertCallbacks(['hello, cruel\n\n'])
        yield self.lbf.flush()
        self.assertCallbacks(['world\n'])

    @defer.inlineCallbacks
    def test_bare_cr_folded(self):
        r"a bare \r is treated as and converted to a newline"
        yield self.lbf.append('1%\r5%\r15%\r100%\nfinished')
        yield self.lbf.flush()
        self.assertCallbacks(['1%\n5%\n15%\n100%\n', 'finished\n'])

    @defer.inlineCallbacks
    def test_backspace_folded(self):
        r"a lot of \b is treated as and converted to a newline"
        yield self.lbf.append('1%\b\b5%\b\b15%\b\b\b100%\nfinished')
        yield self.lbf.flush()
        self.assertCallbacks(['1%\n5%\n15%\n100%\n', 'finished\n'])

    @defer.inlineCallbacks
    def test_mixed_consecutive_newlines(self):
        r"mixing newline styles back-to-back doesn't collapse them"
        yield self.lbf.append('1\r\n\n\r')
        self.assertCallbacks(['1\n\n'])  # last \r is delayed until flush
        yield self.lbf.append('2\n\r\n')
        self.assertCallbacks(['\n2\n\n'])

    @defer.inlineCallbacks
    def test_split_newlines(self):
        r"multi-character newlines, split across chunks, are converted"
        input = 'a\nb\r\nc\rd\n\re'
        for splitpoint in range(1, len(input) - 1):
            a, b = input[:splitpoint], input[splitpoint:]
            yield self.lbf.append(a)
            yield self.lbf.append(b)
            yield self.lbf.flush()
            res = ''.join(self.callbacks)
            log.msg(f'feeding {repr(a)}, {repr(b)} gives {repr(res)}')
            self.assertEqual(res, 'a\nb\nc\nd\n\ne\n')
            self.callbacks = []

    @defer.inlineCallbacks
    def test_split_terminal_control(self):
        """terminal control characters are converted"""
        yield self.lbf.append('1234\033[u4321')
        yield self.lbf.flush()
        self.assertCallbacks(['1234\n', '4321\n'])
        yield self.lbf.append('1234\033[1;2H4321')
        yield self.lbf.flush()
        self.assertCallbacks(['1234\n', '4321\n'])
        yield self.lbf.append('1234\033[1;2f4321')
        yield self.lbf.flush()
        self.assertCallbacks(['1234\n', '4321\n'])

    @defer.inlineCallbacks
    def test_long_lines(self):
        """long lines are split"""
        for _ in range(4):
            yield self.lbf.append('12' * 1000)
        # a split at 4096 + the remaining chars
        self.assertCallbacks(['12' * 2048 + '\n' + '12' * 952 + '\n'])

    @defer.inlineCallbacks
    def test_huge_lines(self):
        """huge lines are split"""
        yield self.lbf.append('12' * 32768)
        yield self.lbf.flush()
        self.assertCallbacks([('12' * 2048 + '\n') * 16])

    @defer.inlineCallbacks
    def test_empty_flush(self):
        yield self.lbf.flush()

        self.assertEqual(self.callbacks, [])


class LBFNoCallback(unittest.TestCase):

    def setUp(self):
        self.lbf = lineboundaries.LineBoundaryFinder()

    def test_already_terminated(self):
        res = self.lbf.append('abcd\ndefg\n')
        self.assertEqual(res, 'abcd\ndefg\n')
        res = self.lbf.append('xyz\n')
        self.assertEqual(res, 'xyz\n')
        res = self.lbf.flush()
        self.assertEqual(res, None)

    def test_partial_line(self):
        res = self.lbf.append('hello\nworld')
        self.assertEqual(res, 'hello\n')
        res = self.lbf.flush()
        self.assertEqual(res, 'world\n')

    def test_empty_appends(self):
        res = self.lbf.append('hello ')
        self.assertEqual(res, None)

        res = self.lbf.append('')
        self.assertEqual(res, None)

        res = self.lbf.append('world\n')
        self.assertEqual(res, 'hello world\n')

        res = self.lbf.append('')
        self.assertEqual(res, None)

    def test_embedded_newlines(self):
        res = self.lbf.append('hello, ')
        self.assertEqual(res, None)

        res = self.lbf.append('cruel\nworld')
        self.assertEqual(res, 'hello, cruel\n')

        res = self.lbf.flush()
        self.assertEqual(res, 'world\n')

    def test_windows_newlines_folded(self):
        r"Windows' \r\n is treated as and converted to a newline"
        res = self.lbf.append('hello, ')
        self.assertEqual(res, None)

        res = self.lbf.append('cruel\r\n\r\nworld')
        self.assertEqual(res, 'hello, cruel\n\n')

        res = self.lbf.flush()
        self.assertEqual(res, 'world\n')

    def test_bare_cr_folded(self):
        r"a bare \r is treated as and converted to a newline"
        self.lbf.append('1%\r5%\r15%\r100%\nfinished')
        res = self.lbf.flush()
        self.assertEqual(res, 'finished\n')

    def test_backspace_folded(self):
        r"a lot of \b is treated as and converted to a newline"
        self.lbf.append('1%\b\b5%\b\b15%\b\b\b100%\nfinished')
        res = self.lbf.flush()
        self.assertEqual(res, 'finished\n')

    def test_mixed_consecutive_newlines(self):
        r"mixing newline styles back-to-back doesn't collapse them"
        res = self.lbf.append('1\r\n\n\r')
        self.assertEqual(res, '1\n\n')

        res = self.lbf.append('2\n\r\n')
        self.assertEqual(res, '\n2\n\n')

    def test_split_newlines(self):
        r"multi-character newlines, split across chunks, are converted"
        input = 'a\nb\r\nc\rd\n\re'
        result = []
        for splitpoint in range(1, len(input) - 1):
            a, b = input[:splitpoint], input[splitpoint:]
            result.append(self.lbf.append(a))
            result.append(self.lbf.append(b))
            result.append(self.lbf.flush())

            result = [e for e in result if e is not None]
            res = ''.join(result)

            log.msg(f'feeding {repr(a)}, {repr(b)} gives {repr(res)}')
            self.assertEqual(res, 'a\nb\nc\nd\n\ne\n')
            result.clear()

    def test_split_terminal_control(self):
        """terminal control characters are converted"""
        res = self.lbf.append('1234\033[u4321')
        self.assertEqual(res, '1234\n')

        res = self.lbf.flush()
        self.assertEqual(res, '4321\n')

        res = self.lbf.append('1234\033[1;2H4321')
        self.assertEqual(res, '1234\n')

        res = self.lbf.flush()
        self.assertEqual(res, '4321\n')

        res = self.lbf.append('1234\033[1;2f4321')
        self.assertEqual(res, '1234\n')

        res = self.lbf.flush()
        self.assertEqual(res, '4321\n')

    def test_long_lines(self):
        """long lines are split"""
        res = []
        for _ in range(4):
            res.append(self.lbf.append('12' * 1000))
        res = [e for e in res if e is not None]
        res = ''.join(res)
        # a split at 4096 + the remaining chars
        self.assertEqual(res, '12' * 2048 + '\n' + '12' * 952 + '\n')

    def test_huge_lines(self):
        """huge lines are split"""
        res = []
        res.append(self.lbf.append('12' * 32768))
        res.append(self.lbf.flush())
        res = [e for e in res if e is not None]
        self.assertEqual(res, [('12' * 2048 + '\n') * 16])

    def test_empty_flush(self):
        res = self.lbf.flush()
        self.assertEqual(res, None)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import json

from autobahn.wamp.exception import TransportLost
from autobahn.wamp.types import PublishOptions
from autobahn.wamp.types import SubscribeOptions
from twisted.internet import defer
from twisted.python import log

from buildbot.mq import base
from buildbot.util import service
from buildbot.util import toJson


class WampMQ(service.ReconfigurableServiceMixin, base.MQBase):
    NAMESPACE = "org.buildbot.mq"

    def produce(self, routingKey, data):
        d = self._produce(routingKey, data)
        d.addErrback(
            log.err, "Problem while producing message on topic " + repr(routingKey))

    @classmethod
    def messageTopic(cls, routingKey):
        def ifNone(v, default):
            return default if v is None else v
        # replace None values by "" in routing key
        routingKey = [ifNone(key, "") for key in routingKey]
        # then join them with "dot", and add the prefix
        return cls.NAMESPACE + "." + ".".join(routingKey)

    @classmethod
    def routingKeyFromMessageTopic(cls, topic):
        # just split the topic, and remove the NAMESPACE prefix
        return tuple(topic[len(WampMQ.NAMESPACE) + 1:].split("."))

    def _produce(self, routingKey, data):
        _data = json.loads(json.dumps(data, default=toJson))
        options = PublishOptions(exclude_me=False)
        return self.master.wamp.publish(self.messageTopic(routingKey), _data, options=options)

    def startConsuming(self, callback, _filter, persistent_name=None):
        if persistent_name is not None:
            log.err(f'wampmq: persistent queues are not persisted: {persistent_name} {_filter}')

        qr = QueueRef(self, callback)

        self._startConsuming(qr, callback, _filter)
        return defer.succeed(qr)

    def _startConsuming(self, qr, callback, _filter, persistent_name=None):
        return qr.subscribe(self.master.wamp, self, _filter)


class QueueRef(base.QueueRef):

    def __init__(self, mq, callback):
        super().__init__(callback)
        self.unreg = None
        self.mq = mq

    @defer.inlineCallbacks
    def subscribe(self, connector_service, wamp_service, _filter):
        self.filter = _filter
        self.emulated = False
        options = dict(details_arg=str('details'))
        if None in _filter:
            options["match"] = "wildcard"
        options = SubscribeOptions(**options)
        _filter = WampMQ.messageTopic(_filter)
        self.unreg = yield connector_service.subscribe(self.wampInvoke, _filter, options=options)
        if self.callback is None:
            yield self.stopConsuming()

    def wampInvoke(self, msg, details):
        if details.topic is not None:
            # in the case of a wildcard, wamp router sends the topic
            topic = WampMQ.routingKeyFromMessageTopic(details.topic)
        else:
            # in the case of an exact match, then we can use our own topic
            topic = self.filter
        self.mq.invokeQref(self, topic, msg)

    @defer.inlineCallbacks
    def stopConsuming(self):
        self.callback = None
        if self.unreg is not None:
            unreg = self.unreg
            self.unreg = None
            try:
                yield unreg.unsubscribe()
            except TransportLost:
                pass
            except Exception as e:
                log.err(e, 'When unsubscribing MQ connection ' + str(unreg))



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import json
import os
import shutil
import sys
from io import StringIO

import mock

import twisted
from twisted.internet import defer
from twisted.protocols import basic
from twisted.trial import unittest

from buildbot.schedulers import trysched
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import dirs
from buildbot.test.util import scheduler


class TryBase(scheduler.SchedulerMixin, TestReactorMixin, unittest.TestCase):
    OBJECTID = 26
    SCHEDULERID = 6

    def setUp(self):
        self.setup_test_reactor()
        self.setUpScheduler()

    def tearDown(self):
        self.tearDownScheduler()

    def makeScheduler(self, **kwargs):
        return self.attachScheduler(trysched.Try_Userpass(**kwargs),
                                    self.OBJECTID, self.SCHEDULERID)

    def test_filterBuilderList_ok(self):
        sched = trysched.TryBase(
            name='tsched', builderNames=['a', 'b', 'c'], properties={})
        self.assertEqual(sched.filterBuilderList(['b', 'c']), ['b', 'c'])

    def test_filterBuilderList_bad(self):
        sched = trysched.TryBase(
            name='tsched', builderNames=['a', 'b'], properties={})
        self.assertEqual(sched.filterBuilderList(['b', 'c']), [])

    def test_filterBuilderList_empty(self):
        sched = trysched.TryBase(
            name='tsched', builderNames=['a', 'b'], properties={})
        self.assertEqual(sched.filterBuilderList([]), ['a', 'b'])

    @defer.inlineCallbacks
    def test_enabled_callback(self):
        sched = self.makeScheduler(name='tsched', builderNames=['a'],
                                   port='tcp:9999', userpass=[('fred', 'derf')])
        expectedValue = not sched.enabled
        yield sched._enabledCallback(None, {'enabled': not sched.enabled})
        self.assertEqual(sched.enabled, expectedValue)
        expectedValue = not sched.enabled
        yield sched._enabledCallback(None, {'enabled': not sched.enabled})
        self.assertEqual(sched.enabled, expectedValue)

    @defer.inlineCallbacks
    def test_disabled_activate(self):
        sched = self.makeScheduler(name='tsched', builderNames=['a'],
                                   port='tcp:9999', userpass=[('fred', 'derf')])
        yield sched._enabledCallback(None, {'enabled': not sched.enabled})
        self.assertEqual(sched.enabled, False)
        r = yield sched.activate()
        self.assertEqual(r, None)

    @defer.inlineCallbacks
    def test_disabled_deactivate(self):
        sched = self.makeScheduler(name='tsched', builderNames=['a'],
                                   port='tcp:9999', userpass=[('fred', 'derf')])
        yield sched._enabledCallback(None, {'enabled': not sched.enabled})
        self.assertEqual(sched.enabled, False)
        r = yield sched.deactivate()
        self.assertEqual(r, None)


class JobdirService(dirs.DirsMixin, unittest.TestCase):

    def setUp(self):
        self.jobdir = 'jobdir'
        self.newdir = os.path.join(self.jobdir, 'new')
        self.curdir = os.path.join(self.jobdir, 'cur')
        self.tmpdir = os.path.join(self.jobdir, 'tmp')
        self.setUpDirs(self.jobdir, self.newdir, self.curdir, self.tmpdir)

    def tearDown(self):
        self.tearDownDirs()

    def test_messageReceived(self):
        # stub out svc.scheduler.handleJobFile and .jobdir
        scheduler = mock.Mock()

        def handleJobFile(filename, f):
            self.assertEqual(filename, 'jobdata')
            self.assertEqual(f.read(), 'JOBDATA')
        scheduler.handleJobFile = handleJobFile
        scheduler.jobdir = self.jobdir

        svc = trysched.JobdirService(scheduler=scheduler, basedir=self.jobdir)

        # create some new data to process
        jobdata = os.path.join(self.newdir, 'jobdata')
        with open(jobdata, "w", encoding='utf-8') as f:
            f.write('JOBDATA')

        # run it
        svc.messageReceived('jobdata')


class Try_Jobdir(scheduler.SchedulerMixin, TestReactorMixin, unittest.TestCase):

    OBJECTID = 23
    SCHEDULERID = 3

    def setUp(self):
        self.setup_test_reactor()
        self.setUpScheduler()
        self.jobdir = None

    def tearDown(self):
        self.tearDownScheduler()
        if self.jobdir:
            shutil.rmtree(self.jobdir)

    # tests

    def setup_test_startService(self, jobdir, exp_jobdir):
        # set up jobdir
        self.jobdir = os.path.abspath('jobdir')
        if os.path.exists(self.jobdir):
            shutil.rmtree(self.jobdir)
        os.mkdir(self.jobdir)

        # build scheduler
        kwargs = dict(name="tsched", builderNames=['a'], jobdir=self.jobdir)
        sched = self.attachScheduler(
            trysched.Try_Jobdir(**kwargs), self.OBJECTID, self.SCHEDULERID,
            overrideBuildsetMethods=True)

        # watch interaction with the watcher service
        sched.watcher.startService = mock.Mock()
        sched.watcher.stopService = mock.Mock()

    @defer.inlineCallbacks
    def do_test_startService(self):
        # start it
        yield self.sched.startService()

        # check that it has set the basedir correctly
        self.assertEqual(self.sched.watcher.basedir, self.jobdir)
        self.assertEqual(1, self.sched.watcher.startService.call_count)
        self.assertEqual(0, self.sched.watcher.stopService.call_count)

        yield self.sched.stopService()

        self.assertEqual(1, self.sched.watcher.startService.call_count)
        self.assertEqual(1, self.sched.watcher.stopService.call_count)

    def test_startService_reldir(self):
        self.setup_test_startService(
            'jobdir',
            os.path.abspath('basedir/jobdir'))
        return self.do_test_startService()

    def test_startService_reldir_subdir(self):
        self.setup_test_startService(
            'jobdir',
            os.path.abspath('basedir/jobdir/cur'))
        return self.do_test_startService()

    def test_startService_absdir(self):
        self.setup_test_startService(
            os.path.abspath('jobdir'),
            os.path.abspath('jobdir'))
        return self.do_test_startService()

    @defer.inlineCallbacks
    def do_test_startService_but_not_active(self, jobdir, exp_jobdir):
        """Same as do_test_startService, but the master wont activate this service"""
        self.setup_test_startService(
            'jobdir',
            os.path.abspath('basedir/jobdir'))

        self.setSchedulerToMaster(self.OTHER_MASTER_ID)

        # start it
        self.sched.startService()

        # check that it has set the basedir correctly, even if it doesn't start
        self.assertEqual(self.sched.watcher.basedir, self.jobdir)

        yield self.sched.stopService()

        self.assertEqual(0, self.sched.watcher.startService.call_count)
        self.assertEqual(0, self.sched.watcher.stopService.call_count)

    # parseJob

    def test_parseJob_empty(self):
        sched = trysched.Try_Jobdir(
            name='tsched', builderNames=['a'], jobdir='foo')
        with self.assertRaises(trysched.BadJobfile):
            sched.parseJob(StringIO(''))

    def test_parseJob_longer_than_netstring_MAXLENGTH(self):
        self.patch(basic.NetstringReceiver, 'MAX_LENGTH', 100)
        sched = trysched.Try_Jobdir(
            name='tsched', builderNames=['a'], jobdir='foo')
        jobstr = self.makeNetstring(
            '1', 'extid', 'trunk', '1234', '1', 'this is my diff, -- ++, etc.',
            'buildera', 'builderc'
        )
        jobstr += 'x' * 200

        test_temp_file = StringIO(jobstr)

        with self.assertRaises(trysched.BadJobfile):
            sched.parseJob(test_temp_file)

    def test_parseJob_invalid(self):
        sched = trysched.Try_Jobdir(
            name='tsched', builderNames=['a'], jobdir='foo')
        with self.assertRaises(trysched.BadJobfile):
            sched.parseJob(StringIO('this is not a netstring'))

    def test_parseJob_invalid_version(self):
        sched = trysched.Try_Jobdir(
            name='tsched', builderNames=['a'], jobdir='foo')
        with self.assertRaises(trysched.BadJobfile):
            sched.parseJob(StringIO('1:9,'))

    def makeNetstring(self, *strings):
        return ''.join([f'{len(s)}:{s},' for s in strings])

    def test_parseJob_v1(self):
        sched = trysched.Try_Jobdir(
            name='tsched', builderNames=['buildera', 'builderb'], jobdir='foo')
        jobstr = self.makeNetstring(
            '1', 'extid', 'trunk', '1234', '1', 'this is my diff, -- ++, etc.',
            'buildera', 'builderc'
        )
        parsedjob = sched.parseJob(StringIO(jobstr))
        self.assertEqual(parsedjob, {
            'baserev': '1234',
            'branch': 'trunk',
            'builderNames': ['buildera', 'builderc'],
            'jobid': 'extid',
            'patch_body': b'this is my diff, -- ++, etc.',
            'patch_level': 1,
            'project': '',
            'who': '',
            'comment': '',
            'repository': '',
            'properties': {},
        })

    def test_parseJob_v1_empty_branch_rev(self):
        sched = trysched.Try_Jobdir(
            name='tsched', builderNames=['buildera', 'builderb'], jobdir='foo')
        jobstr = self.makeNetstring(
            # blank branch, rev are turned to None
            '1', 'extid', '', '', '1', 'this is my diff, -- ++, etc.',
            'buildera', 'builderc'
        )
        parsedjob = sched.parseJob(StringIO(jobstr))
        self.assertEqual(parsedjob['branch'], None)
        self.assertEqual(parsedjob['baserev'], None)

    def test_parseJob_v1_no_builders(self):
        sched = trysched.Try_Jobdir(
            name='tsched', builderNames=['buildera', 'builderb'], jobdir='foo')
        jobstr = self.makeNetstring(
            '1', 'extid', '', '', '1', 'this is my diff, -- ++, etc.'
        )
        parsedjob = sched.parseJob(StringIO(jobstr))
        self.assertEqual(parsedjob['builderNames'], [])

    def test_parseJob_v1_no_properties(self):
        sched = trysched.Try_Jobdir(
            name='tsched', builderNames=['buildera', 'builderb'], jobdir='foo')
        jobstr = self.makeNetstring(
            '1', 'extid', '', '', '1', 'this is my diff, -- ++, etc.'
        )
        parsedjob = sched.parseJob(StringIO(jobstr))
        self.assertEqual(parsedjob['properties'], {})

    def test_parseJob_v2(self):
        sched = trysched.Try_Jobdir(
            name='tsched', builderNames=['buildera', 'builderb'], jobdir='foo')
        jobstr = self.makeNetstring(
            '2', 'extid', 'trunk', '1234', '1', 'this is my diff, -- ++, etc.',
            'repo', 'proj',
            'buildera', 'builderc'
        )
        parsedjob = sched.parseJob(StringIO(jobstr))
        self.assertEqual(parsedjob, {
            'baserev': '1234',
            'branch': 'trunk',
            'builderNames': ['buildera', 'builderc'],
            'jobid': 'extid',
            'patch_body': b'this is my diff, -- ++, etc.',
            'patch_level': 1,
            'project': 'proj',
            'who': '',
            'comment': '',
            'repository': 'repo',
            'properties': {},
        })

    def test_parseJob_v2_empty_branch_rev(self):
        sched = trysched.Try_Jobdir(
            name='tsched', builderNames=['buildera', 'builderb'], jobdir='foo')
        jobstr = self.makeNetstring(
            # blank branch, rev are turned to None
            '2', 'extid', '', '', '1', 'this is my diff, -- ++, etc.',
            'repo', 'proj',
            'buildera', 'builderc'
        )
        parsedjob = sched.parseJob(StringIO(jobstr))
        self.assertEqual(parsedjob['branch'], None)
        self.assertEqual(parsedjob['baserev'], None)

    def test_parseJob_v2_no_builders(self):
        sched = trysched.Try_Jobdir(
            name='tsched', builderNames=['buildera', 'builderb'], jobdir='foo')
        jobstr = self.makeNetstring(
            '2', 'extid', 'trunk', '1234', '1', 'this is my diff, -- ++, etc.',
            'repo', 'proj',
        )
        parsedjob = sched.parseJob(StringIO(jobstr))
        self.assertEqual(parsedjob['builderNames'], [])

    def test_parseJob_v2_no_properties(self):
        sched = trysched.Try_Jobdir(
            name='tsched', builderNames=['buildera', 'builderb'], jobdir='foo')
        jobstr = self.makeNetstring(
            '2', 'extid', 'trunk', '1234', '1', 'this is my diff, -- ++, etc.',
            'repo', 'proj',
        )
        parsedjob = sched.parseJob(StringIO(jobstr))
        self.assertEqual(parsedjob['properties'], {})

    def test_parseJob_v3(self):
        sched = trysched.Try_Jobdir(
            name='tsched', builderNames=['buildera', 'builderb'], jobdir='foo')
        jobstr = self.makeNetstring(
            '3', 'extid', 'trunk', '1234', '1', 'this is my diff, -- ++, etc.',
            'repo', 'proj', 'who',
            'buildera', 'builderc'
        )
        parsedjob = sched.parseJob(StringIO(jobstr))
        self.assertEqual(parsedjob, {
            'baserev': '1234',
            'branch': 'trunk',
            'builderNames': ['buildera', 'builderc'],
            'jobid': 'extid',
            'patch_body': b'this is my diff, -- ++, etc.',
            'patch_level': 1,
            'project': 'proj',
            'who': 'who',
            'comment': '',
            'repository': 'repo',
            'properties': {},
        })

    def test_parseJob_v3_empty_branch_rev(self):
        sched = trysched.Try_Jobdir(
            name='tsched', builderNames=['buildera', 'builderb'], jobdir='foo')
        jobstr = self.makeNetstring(
            # blank branch, rev are turned to None
            '3', 'extid', '', '', '1', 'this is my diff, -- ++, etc.',
            'repo', 'proj', 'who',
            'buildera', 'builderc'
        )
        parsedjob = sched.parseJob(StringIO(jobstr))
        self.assertEqual(parsedjob['branch'], None)
        self.assertEqual(parsedjob['baserev'], None)

    def test_parseJob_v3_no_builders(self):
        sched = trysched.Try_Jobdir(
            name='tsched', builderNames=['buildera', 'builderb'], jobdir='foo')
        jobstr = self.makeNetstring(
            '3', 'extid', 'trunk', '1234', '1', 'this is my diff, -- ++, etc.',
            'repo', 'proj', 'who'
        )
        parsedjob = sched.parseJob(StringIO(jobstr))
        self.assertEqual(parsedjob['builderNames'], [])

    def test_parseJob_v3_no_properties(self):
        sched = trysched.Try_Jobdir(
            name='tsched', builderNames=['buildera', 'builderb'], jobdir='foo')
        jobstr = self.makeNetstring(
            '3', 'extid', 'trunk', '1234', '1', 'this is my diff, -- ++, etc.',
            'repo', 'proj', 'who'
        )
        parsedjob = sched.parseJob(StringIO(jobstr))
        self.assertEqual(parsedjob['properties'], {})

    def test_parseJob_v4(self):
        sched = trysched.Try_Jobdir(
            name='tsched', builderNames=['buildera', 'builderb'], jobdir='foo')
        jobstr = self.makeNetstring(
            '4', 'extid', 'trunk', '1234', '1', 'this is my diff, -- ++, etc.',
            'repo', 'proj', 'who', 'comment',
            'buildera', 'builderc'
        )
        parsedjob = sched.parseJob(StringIO(jobstr))
        self.assertEqual(parsedjob, {
            'baserev': '1234',
            'branch': 'trunk',
            'builderNames': ['buildera', 'builderc'],
            'jobid': 'extid',
            'patch_body': b'this is my diff, -- ++, etc.',
            'patch_level': 1,
            'project': 'proj',
            'who': 'who',
            'comment': 'comment',
            'repository': 'repo',
            'properties': {},
        })

    def test_parseJob_v4_empty_branch_rev(self):
        sched = trysched.Try_Jobdir(
            name='tsched', builderNames=['buildera', 'builderb'], jobdir='foo')
        jobstr = self.makeNetstring(
            # blank branch, rev are turned to None
            '4', 'extid', '', '', '1', 'this is my diff, -- ++, etc.',
            'repo', 'proj', 'who', 'comment',
            'buildera', 'builderc'
        )
        parsedjob = sched.parseJob(StringIO(jobstr))
        self.assertEqual(parsedjob['branch'], None)
        self.assertEqual(parsedjob['baserev'], None)

    def test_parseJob_v4_no_builders(self):
        sched = trysched.Try_Jobdir(
            name='tsched', builderNames=['buildera', 'builderb'], jobdir='foo')
        jobstr = self.makeNetstring(
            '4', 'extid', 'trunk', '1234', '1', 'this is my diff, -- ++, etc.',
            'repo', 'proj', 'who', 'comment'
        )
        parsedjob = sched.parseJob(StringIO(jobstr))
        self.assertEqual(parsedjob['builderNames'], [])

    def test_parseJob_v4_no_properties(self):
        sched = trysched.Try_Jobdir(
            name='tsched', builderNames=['buildera', 'builderb'], jobdir='foo')
        jobstr = self.makeNetstring(
            '4', 'extid', 'trunk', '1234', '1', 'this is my diff, -- ++, etc.',
            'repo', 'proj', 'who', 'comment'
        )
        parsedjob = sched.parseJob(StringIO(jobstr))
        self.assertEqual(parsedjob['properties'], {})

    def test_parseJob_v5(self):
        sched = trysched.Try_Jobdir(
            name='tsched', builderNames=['buildera', 'builderb'], jobdir='foo')
        jobstr = self.makeNetstring(
            '5',
            json.dumps({
                'jobid': 'extid', 'branch': 'trunk', 'baserev': '1234',
                'patch_level': 1, 'patch_body': 'this is my diff, -- ++, etc.',
                'repository': 'repo', 'project': 'proj', 'who': 'who',
                'comment': 'comment', 'builderNames': ['buildera', 'builderc'],
                'properties': {'foo': 'bar'},
            }))
        parsedjob = sched.parseJob(StringIO(jobstr))
        self.assertEqual(parsedjob, {
            'baserev': '1234',
            'branch': 'trunk',
            'builderNames': ['buildera', 'builderc'],
            'jobid': 'extid',
            'patch_body': b'this is my diff, -- ++, etc.',
            'patch_level': 1,
            'project': 'proj',
            'who': 'who',
            'comment': 'comment',
            'repository': 'repo',
            'properties': {'foo': 'bar'},
        })

    def test_parseJob_v5_empty_branch_rev(self):
        sched = trysched.Try_Jobdir(
            name='tsched', builderNames=['buildera', 'builderb'], jobdir='foo')
        jobstr = self.makeNetstring(
            # blank branch, rev are turned to None
            '4', 'extid', '', '', '1', 'this is my diff, -- ++, etc.',
            'repo', 'proj', 'who', 'comment',
            'buildera', 'builderc'
        )
        parsedjob = sched.parseJob(StringIO(jobstr))
        self.assertEqual(parsedjob['branch'], None)
        self.assertEqual(parsedjob['baserev'], None)

    def test_parseJob_v5_no_builders(self):
        sched = trysched.Try_Jobdir(
            name='tsched', builderNames=['buildera', 'builderb'], jobdir='foo')
        jobstr = self.makeNetstring(
            '5',
            json.dumps({
                'jobid': 'extid', 'branch': 'trunk', 'baserev': '1234',
                'patch_level': '1', 'patch_body': 'this is my diff, -- ++, etc.',
                'repository': 'repo', 'project': 'proj', 'who': 'who',
                'comment': 'comment', 'builderNames': [],
                'properties': {'foo': 'bar'},
            }))
        parsedjob = sched.parseJob(StringIO(jobstr))
        self.assertEqual(parsedjob['builderNames'], [])

    def test_parseJob_v5_no_properties(self):
        sched = trysched.Try_Jobdir(
            name='tsched', builderNames=['buildera', 'builderb'], jobdir='foo')
        jobstr = self.makeNetstring(
            '5',
            json.dumps({
                'jobid': 'extid', 'branch': 'trunk', 'baserev': '1234',
                'patch_level': '1', 'patch_body': 'this is my diff, -- ++, etc.',
                'repository': 'repo', 'project': 'proj', 'who': 'who',
                'comment': 'comment', 'builderNames': ['buildera', 'builderb'],
                'properties': {},
            }))
        parsedjob = sched.parseJob(StringIO(jobstr))
        self.assertEqual(parsedjob['properties'], {})

    def test_parseJob_v5_invalid_json(self):
        sched = trysched.Try_Jobdir(
            name='tsched', builderNames=['buildera', 'builderb'], jobdir='foo')
        jobstr = self.makeNetstring('5', '{"comment": "com}')
        with self.assertRaises(trysched.BadJobfile):
            sched.parseJob(StringIO(jobstr))

    # handleJobFile

    def call_handleJobFile(self, parseJob):
        sched = self.attachScheduler(
            trysched.Try_Jobdir(
                name='tsched', builderNames=['buildera', 'builderb'],
                jobdir='foo'), self.OBJECTID, self.SCHEDULERID,
            overrideBuildsetMethods=True,
            createBuilderDB=True)
        fakefile = mock.Mock()

        def parseJob_(f):
            assert f is fakefile
            return parseJob(f)
        sched.parseJob = parseJob_
        return defer.maybeDeferred(sched.handleJobFile, 'fakefile', fakefile)

    def makeSampleParsedJob(self, **overrides):
        pj = dict(baserev='1234', branch='trunk',
                  builderNames=['buildera', 'builderb'],
                  jobid='extid', patch_body=b'this is my diff, -- ++, etc.',
                  patch_level=1, project='proj', repository='repo', who='who',
                  comment='comment', properties={})
        pj.update(overrides)
        return pj

    @defer.inlineCallbacks
    def test_handleJobFile(self):
        yield self.call_handleJobFile(lambda f: self.makeSampleParsedJob())

        self.assertEqual(self.addBuildsetCalls, [
            ('addBuildsetForSourceStamps', dict(
                builderNames=['buildera', 'builderb'],
                external_idstring='extid',
                properties={},
                reason="'try' job by user who",
                sourcestamps=[
                    dict(
                        branch='trunk',
                        codebase='',
                        patch_author='who',
                        patch_body=b'this is my diff, -- ++, etc.',
                        patch_comment='comment',
                        patch_level=1,
                        patch_subdir='',
                        project='proj',
                        repository='repo',
                        revision='1234'),
                ])),
        ])

    @defer.inlineCallbacks
    def test_handleJobFile_exception(self):
        def parseJob(f):
            raise trysched.BadJobfile
        yield self.call_handleJobFile(parseJob)

        self.assertEqual(self.addBuildsetCalls, [])
        self.assertEqual(
            1, len(self.flushLoggedErrors(trysched.BadJobfile)))
    if twisted.version.major <= 9 and sys.version_info[:2] >= (2, 7):
        test_handleJobFile_exception.skip = (
            "flushLoggedErrors does not work correctly on 9.0.0 "
            "and earlier with Python-2.7")

    @defer.inlineCallbacks
    def test_handleJobFile_bad_builders(self):
        yield self.call_handleJobFile(
            lambda f: self.makeSampleParsedJob(builderNames=['xxx']))

        self.assertEqual(self.addBuildsetCalls, [])

    @defer.inlineCallbacks
    def test_handleJobFile_subset_builders(self):
        yield self.call_handleJobFile(
            lambda f: self.makeSampleParsedJob(builderNames=['buildera']))

        self.assertEqual(self.addBuildsetCalls, [
            ('addBuildsetForSourceStamps', dict(
                builderNames=['buildera'],
                external_idstring='extid',
                properties={},
                reason="'try' job by user who",
                sourcestamps=[
                    dict(
                        branch='trunk',
                        codebase='',
                        patch_author='who',
                        patch_body=b'this is my diff, -- ++, etc.',
                        patch_comment='comment',
                        patch_level=1,
                        patch_subdir='',
                        project='proj',
                        repository='repo',
                        revision='1234'),
                ])),
        ])

    @defer.inlineCallbacks
    def test_handleJobFile_with_try_properties(self):
        yield self.call_handleJobFile(
                lambda f: self.makeSampleParsedJob(properties={'foo': 'bar'}))

        self.assertEqual(self.addBuildsetCalls, [
            ('addBuildsetForSourceStamps', dict(
                builderNames=['buildera', 'builderb'],
                external_idstring='extid',
                properties={'foo': ('bar', 'try build')},
                reason="'try' job by user who",
                sourcestamps=[
                    dict(
                        branch='trunk',
                        codebase='',
                        patch_author='who',
                        patch_body=b'this is my diff, -- ++, etc.',
                        patch_comment='comment',
                        patch_level=1,
                        patch_subdir='',
                        project='proj',
                        repository='repo',
                        revision='1234'),
                ])),
        ])

    def test_handleJobFile_with_invalid_try_properties(self):
        d = self.call_handleJobFile(
            lambda f: self.makeSampleParsedJob(properties=['foo', 'bar']))
        return self.assertFailure(d, AttributeError)


class Try_Userpass_Perspective(scheduler.SchedulerMixin, TestReactorMixin,
                               unittest.TestCase):

    OBJECTID = 26
    SCHEDULERID = 6

    def setUp(self):
        self.setup_test_reactor()
        self.setUpScheduler()

    def tearDown(self):
        self.tearDownScheduler()

    def makeScheduler(self, **kwargs):
        return self.attachScheduler(trysched.Try_Userpass(**kwargs),
                                    self.OBJECTID, self.SCHEDULERID,
                                    overrideBuildsetMethods=True,
                                    createBuilderDB=True)

    @defer.inlineCallbacks
    def call_perspective_try(self, *args, **kwargs):
        sched = self.makeScheduler(name='tsched', builderNames=['a', 'b'],
                                   port='xxx', userpass=[('a', 'b')],
                                   properties=dict(frm='schd'))
        persp = trysched.Try_Userpass_Perspective(sched, 'a')

        # patch out all of the handling after addBuildsetForSourceStamp
        def getBuildset(bsid):
            return dict(bsid=bsid)
        self.db.buildsets.getBuildset = getBuildset

        rbss = yield persp.perspective_try(*args, **kwargs)

        if rbss is None:
            return
        self.assertIsInstance(rbss, trysched.RemoteBuildSetStatus)

    @defer.inlineCallbacks
    def test_perspective_try(self):
        yield self.call_perspective_try(
            'default', 'abcdef', (1, '-- ++'), 'repo', 'proj', ['a'],
            properties={'pr': 'op'})

        self.maxDiff = None
        self.assertEqual(self.addBuildsetCalls, [
            ('addBuildsetForSourceStamps', dict(
                builderNames=['a'],
                external_idstring=None,
                properties={'pr': ('op', 'try build')},
                reason="'try' job",
                sourcestamps=[
                    dict(
                        branch='default',
                        codebase='',
                        patch_author='',
                        patch_body=b'-- ++',
                        patch_comment='',
                        patch_level=1,
                        patch_subdir='',
                        project='proj',
                        repository='repo',
                        revision='abcdef'),
                ])),
        ])

    @defer.inlineCallbacks
    def test_perspective_try_bytes(self):
        yield self.call_perspective_try(
            'default', 'abcdef', (1, b'-- ++\xf8'), 'repo', 'proj', ['a'],
            properties={'pr': 'op'})

        self.assertEqual(self.addBuildsetCalls, [
            ('addBuildsetForSourceStamps', {
                'builderNames': ['a'],
                'external_idstring': None,
                'properties': {'pr': ('op', 'try build')},
                'reason': "'try' job",
                'sourcestamps': [
                    {
                        'branch': 'default',
                        'codebase': '',
                        'patch_author': '',
                        'patch_body': b'-- ++\xf8',
                        'patch_comment': '',
                        'patch_level': 1,
                        'patch_subdir': '',
                        'project': 'proj',
                        'repository': 'repo',
                        'revision': 'abcdef',
                    }
                ]
            }),
        ])

    @defer.inlineCallbacks
    def test_perspective_try_who(self):
        yield self.call_perspective_try(
                'default', 'abcdef', (1, '-- ++'), 'repo', 'proj', ['a'],
                who='who', comment='comment', properties={'pr': 'op'})

        self.assertEqual(self.addBuildsetCalls, [
            ('addBuildsetForSourceStamps', dict(
                builderNames=['a'],
                external_idstring=None,
                properties={'pr': ('op', 'try build')},
                reason="'try' job by user who (comment)",
                sourcestamps=[
                    dict(
                        branch='default',
                        codebase='',
                        patch_author='who',
                        patch_body=b'-- ++',
                        patch_comment='comment',
                        patch_level=1,
                        patch_subdir='',
                        project='proj',
                        repository='repo',
                        revision='abcdef'),
                ])),
        ])

    @defer.inlineCallbacks
    def test_perspective_try_bad_builders(self):
        yield self.call_perspective_try(
                'default', 'abcdef', (1, '-- ++'), 'repo', 'proj', ['xxx'],
                properties={'pr': 'op'})

        self.assertEqual(self.addBuildsetCalls, [])

    @defer.inlineCallbacks
    def test_getAvailableBuilderNames(self):
        sched = self.makeScheduler(name='tsched', builderNames=['a', 'b'],
                                   port='xxx', userpass=[('a', 'b')])
        persp = trysched.Try_Userpass_Perspective(sched, 'a')
        buildernames = yield persp.perspective_getAvailableBuilderNames()

        self.assertEqual(buildernames, ['a', 'b'])


class Try_Userpass(scheduler.SchedulerMixin, TestReactorMixin,
                   unittest.TestCase):

    OBJECTID = 25
    SCHEDULERID = 5

    def setUp(self):
        self.setup_test_reactor()
        self.setUpScheduler()

    def tearDown(self):
        self.tearDownScheduler()

    def makeScheduler(self, **kwargs):
        sched = self.attachScheduler(trysched.Try_Userpass(**kwargs),
                                     self.OBJECTID, self.SCHEDULERID)
        return sched

    @defer.inlineCallbacks
    def test_service(self):
        sched = self.makeScheduler(name='tsched', builderNames=['a'],
                                   port='tcp:9999', userpass=[('fred', 'derf')])
        # patch out the pbmanager's 'register' command both to be sure
        # the registration is correct and to get a copy of the factory
        registration = mock.Mock()
        registration.unregister = lambda: defer.succeed(None)
        sched.master.pbmanager = mock.Mock()

        def register(portstr, user, passwd, factory):
            self.assertEqual([portstr, user, passwd],
                             ['tcp:9999', 'fred', 'derf'])
            self.got_factory = factory
            return defer.succeed(registration)

        sched.master.pbmanager.register = register
        # start it
        yield sched.startService()
        # make a fake connection by invoking the factory, and check that we
        # get the correct perspective
        persp = self.got_factory(mock.Mock(), 'fred')
        self.assertTrue(isinstance(persp, trysched.Try_Userpass_Perspective))
        yield sched.stopService()

    @defer.inlineCallbacks
    def test_service_but_not_active(self):
        sched = self.makeScheduler(name='tsched', builderNames=['a'],
                                   port='tcp:9999', userpass=[('fred', 'derf')])

        self.setSchedulerToMaster(self.OTHER_MASTER_ID)

        sched.master.pbmanager = mock.Mock()

        sched.startService()
        yield sched.stopService()

        self.assertFalse(sched.master.pbmanager.register.called)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
"""
Source step code for Monotone
"""


from twisted.internet import defer
from twisted.internet import reactor
from twisted.python import log

from buildbot.config import ConfigErrors
from buildbot.interfaces import WorkerSetupError
from buildbot.process import buildstep
from buildbot.process import remotecommand
from buildbot.process.results import SUCCESS
from buildbot.steps.source.base import Source


class Monotone(Source):

    """ Class for Monotone with all smarts """

    name = 'monotone'

    renderables = ['repourl']
    possible_methods = ('clobber', 'copy', 'fresh', 'clean')

    def __init__(self, repourl=None, branch=None, progress=False,
                 mode='incremental', method=None, **kwargs):

        self.repourl = repourl
        self.method = method
        self.mode = mode
        self.branch = branch
        self.sourcedata = f"{self.repourl}?{self.branch}"
        self.database = 'db.mtn'
        self.progress = progress
        super().__init__(**kwargs)
        errors = []

        if not self._hasAttrGroupMember('mode', self.mode):
            errors.append(f"mode {self.mode} is not one of {self._listAttrGroupMembers('mode')}")
        if self.mode == 'incremental' and self.method:
            errors.append("Incremental mode does not require method")

        if self.mode == 'full':
            if self.method is None:
                self.method = 'copy'
            elif self.method not in self.possible_methods:
                errors.append(f"Invalid method for mode == {self.mode}")

        if repourl is None:
            errors.append("you must provide repourl")

        if branch is None:
            errors.append("you must provide branch")

        if errors:
            raise ConfigErrors(errors)

    @defer.inlineCallbacks
    def run_vc(self, branch, revision, patch):
        self.revision = revision
        self.stdio_log = yield self.addLogForRemoteCommands("stdio")

        try:
            monotoneInstalled = yield self.checkMonotone()
            if not monotoneInstalled:
                raise WorkerSetupError("Monotone is not installed on worker")

            yield self._checkDb()
            yield self._retryPull()

            # If we're not throwing away the workdir, check if it's
            # somehow patched or modified and revert.
            if self.mode != 'full' or self.method not in ('clobber', 'copy'):
                patched = yield self.sourcedirIsPatched()
                if patched:
                    yield self.clean()

            # Call a mode specific method
            fn = self._getAttrGroupMember('mode', self.mode)
            yield fn()

            if patch:
                yield self.patch(patch)
            yield self.parseGotRevision()
            return SUCCESS
        finally:
            pass  # FIXME: remove this try:raise block

    @defer.inlineCallbacks
    def mode_full(self):
        if self.method == 'clobber':
            yield self.clobber()
            return
        elif self.method == 'copy':
            yield self.copy()
            return

        updatable = yield self._sourcedirIsUpdatable()
        if not updatable:
            yield self.clobber()
        elif self.method == 'clean':
            yield self.clean()
            yield self._update()
        elif self.method == 'fresh':
            yield self.clean(False)
            yield self._update()
        else:
            raise ValueError("Unknown method, check your configuration")

    @defer.inlineCallbacks
    def mode_incremental(self):
        updatable = yield self._sourcedirIsUpdatable()
        if not updatable:
            yield self.clobber()
        else:
            yield self._update()

    @defer.inlineCallbacks
    def clobber(self):
        yield self.runRmdir(self.workdir)
        yield self._checkout()

    @defer.inlineCallbacks
    def copy(self):
        cmd = remotecommand.RemoteCommand('rmdir', {
            'dir': self.workdir,
            'logEnviron': self.logEnviron,
            'timeout': self.timeout, })
        cmd.useLog(self.stdio_log, False)
        yield self.runCommand(cmd)

        self.workdir = 'source'
        yield self.mode_incremental()
        cmd = remotecommand.RemoteCommand('cpdir',
                                          {'fromdir': 'source',
                                           'todir': 'build',
                                           'logEnviron': self.logEnviron,
                                           'timeout': self.timeout, })
        cmd.useLog(self.stdio_log, False)
        yield self.runCommand(cmd)

        self.workdir = 'build'
        return 0

    @defer.inlineCallbacks
    def checkMonotone(self):
        cmd = remotecommand.RemoteShellCommand(self.workdir,
                                               ['mtn', '--version'],
                                               env=self.env,
                                               logEnviron=self.logEnviron,
                                               timeout=self.timeout)
        cmd.useLog(self.stdio_log, False)
        yield self.runCommand(cmd)
        return cmd.rc == 0

    @defer.inlineCallbacks
    def clean(self, ignore_ignored=True):
        files = []
        commands = [['mtn', 'ls', 'unknown']]
        if not ignore_ignored:
            commands.append(['mtn', 'ls', 'ignored'])
        for cmd in commands:
            stdout = yield self._dovccmd(cmd, workdir=self.workdir,
                                         collectStdout=True)
            if not stdout:
                continue
            for filename in stdout.strip().split('\n'):
                filename = self.workdir + '/' + str(filename)
                files.append(filename)

        if not files:
            rc = 0
        else:
            if self.workerVersionIsOlderThan('rmdir', '2.14'):
                rc = yield self.removeFiles(files)
            else:
                rc = yield self.runRmdir(files, abandonOnFailure=False)

        if rc != 0:
            log.msg("Failed removing files")
            raise buildstep.BuildStepFailed()

    @defer.inlineCallbacks
    def removeFiles(self, files):
        for filename in files:
            res = yield self.runRmdir(filename, abandonOnFailure=False)
            if res:
                return res
        return 0

    def _checkout(self, abandonOnFailure=False):
        command = ['mtn', 'checkout', self.workdir, '--db', self.database]
        if self.revision:
            command.extend(['--revision', self.revision])
        command.extend(['--branch', self.branch])
        return self._dovccmd(command, workdir='.',
                             abandonOnFailure=abandonOnFailure)

    def _update(self, abandonOnFailure=False):
        command = ['mtn', 'update']
        if self.revision:
            command.extend(['--revision', self.revision])
        else:
            command.extend(['--revision', 'h:' + self.branch])
        command.extend(['--branch', self.branch])
        return self._dovccmd(command, workdir=self.workdir,
                             abandonOnFailure=abandonOnFailure)

    def _pull(self, abandonOnFailure=False):
        command = ['mtn', 'pull', self.sourcedata, '--db', self.database]
        if self.progress:
            command.extend(['--ticker=dot'])
        else:
            command.extend(['--ticker=none'])
        d = self._dovccmd(command, workdir='.',
                          abandonOnFailure=abandonOnFailure)
        return d

    @defer.inlineCallbacks
    def _retryPull(self):
        if self.retry:
            abandonOnFailure = (self.retry[1] <= 0)
        else:
            abandonOnFailure = True

        res = yield self._pull(abandonOnFailure)
        if self.retry:
            delay, repeats = self.retry
            if self.stopped or res == 0 or repeats <= 0:
                return res
            else:
                log.msg(f"Checkout failed, trying {repeats} more times after {delay} seconds")
                self.retry = (delay, repeats - 1)
                df = defer.Deferred()
                df.addCallback(lambda _: self._retryPull())
                reactor.callLater(delay, df.callback, None)
                yield df
        return None

    @defer.inlineCallbacks
    def parseGotRevision(self):
        stdout = yield self._dovccmd(['mtn', 'automate', 'select', 'w:'],
                                     workdir=self.workdir,
                                     collectStdout=True)
        revision = stdout.strip()
        if len(revision) != 40:
            raise buildstep.BuildStepFailed()
        log.msg(f"Got Monotone revision {revision}")
        self.updateSourceProperty('got_revision', revision)
        return 0

    @defer.inlineCallbacks
    def _dovccmd(self, command, workdir,
                 collectStdout=False, initialStdin=None, decodeRC=None,
                 abandonOnFailure=True):
        if not command:
            raise ValueError("No command specified")

        if decodeRC is None:
            decodeRC = {0: SUCCESS}
        cmd = remotecommand.RemoteShellCommand(workdir, command,
                                               env=self.env,
                                               logEnviron=self.logEnviron,
                                               timeout=self.timeout,
                                               collectStdout=collectStdout,
                                               initialStdin=initialStdin,
                                               decodeRC=decodeRC)
        cmd.useLog(self.stdio_log, False)
        yield self.runCommand(cmd)

        if abandonOnFailure and cmd.didFail():
            log.msg(f"Source step failed while running command {cmd}")
            raise buildstep.BuildStepFailed()
        if collectStdout:
            return cmd.stdout
        else:
            return cmd.rc

    @defer.inlineCallbacks
    def _checkDb(self):
        db_exists = yield self.pathExists(self.database)
        db_needs_init = False
        if db_exists:
            stdout = yield self._dovccmd(
                ['mtn', 'db', 'info', '--db', self.database],
                workdir='.',
                collectStdout=True)
            if stdout.find("migration needed") >= 0:
                log.msg("Older format database found, migrating it")
                yield self._dovccmd(['mtn', 'db', 'migrate', '--db',
                                     self.database],
                                    workdir='.')
            elif stdout.find("too new, cannot use") >= 0 or \
                    stdout.find("database has no tables") >= 0:
                # The database is of a newer format which the worker's
                # mtn version can not handle. Drop it and pull again
                # with that monotone version installed on the
                # worker. Do the same if it's an empty file.
                yield self.runRmdir(self.database)
                db_needs_init = True
            elif stdout.find("not a monotone database") >= 0:
                # There exists a database file, but it's not a valid
                # monotone database. Do not delete it, but fail with
                # an error.
                raise buildstep.BuildStepFailed()
            else:
                log.msg("Database exists and compatible")
        else:
            db_needs_init = True
            log.msg("Database does not exist")

        if db_needs_init:
            command = ['mtn', 'db', 'init', '--db', self.database]
            yield self._dovccmd(command, workdir='.')

    @defer.inlineCallbacks
    def _sourcedirIsUpdatable(self):
        workdir_path = self.build.path_module.join(self.workdir, '_MTN')
        workdir_exists = yield self.pathExists(workdir_path)

        if not workdir_exists:
            log.msg("Workdir does not exist, falling back to a fresh clone")

        return workdir_exists



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import json

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.config.master import MasterConfig
from buildbot.schedulers.forcesched import AnyPropertyParameter
from buildbot.schedulers.forcesched import BaseParameter
from buildbot.schedulers.forcesched import BooleanParameter
from buildbot.schedulers.forcesched import ChoiceStringParameter
from buildbot.schedulers.forcesched import CodebaseParameter
from buildbot.schedulers.forcesched import CollectedValidationError
from buildbot.schedulers.forcesched import FileParameter
from buildbot.schedulers.forcesched import FixedParameter
from buildbot.schedulers.forcesched import ForceScheduler
from buildbot.schedulers.forcesched import IntParameter
from buildbot.schedulers.forcesched import NestedParameter
from buildbot.schedulers.forcesched import PatchParameter
from buildbot.schedulers.forcesched import StringParameter
from buildbot.schedulers.forcesched import UserNameParameter
from buildbot.schedulers.forcesched import oneCodebase
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import scheduler
from buildbot.test.util.config import ConfigErrorsMixin


class TestForceScheduler(scheduler.SchedulerMixin, ConfigErrorsMixin,
                         TestReactorMixin, unittest.TestCase):

    OBJECTID = 19
    SCHEDULERID = 9
    maxDiff = None

    def setUp(self):
        self.setup_test_reactor()
        self.setUpScheduler()

    def tearDown(self):
        self.tearDownScheduler()

    def makeScheduler(self, name='testsched', builderNames=None,
                      **kw):
        if builderNames is None:
            builderNames = ['a', 'b']
        sched = self.attachScheduler(
            ForceScheduler(name=name, builderNames=builderNames, **kw),
            self.OBJECTID, self.SCHEDULERID,
            overrideBuildsetMethods=True,
            createBuilderDB=True)
        sched.master.config = MasterConfig()

        self.assertEqual(sched.name, name)

        return sched

    # tests

    def test_compare_branch(self):
        self.assertNotEqual(
            ForceScheduler(name="testched", builderNames=[]),
            ForceScheduler(
                name="testched", builderNames=[],
                codebases=oneCodebase(
                    branch=FixedParameter("branch", "fishing/pole"))))

    def test_compare_reason(self):
        self.assertNotEqual(
            ForceScheduler(name="testched", builderNames=[],
                           reason=FixedParameter("reason", "no fish for you!")),
            ForceScheduler(name="testched", builderNames=[],
                           reason=FixedParameter("reason", "thanks for the fish!")))

    def test_compare_revision(self):
        self.assertNotEqual(
            ForceScheduler(
                name="testched", builderNames=[],
                codebases=oneCodebase(
                    revision=FixedParameter("revision", "fish-v1"))),
            ForceScheduler(
                name="testched", builderNames=[],
                codebases=oneCodebase(
                    revision=FixedParameter("revision", "fish-v2"))))

    def test_compare_repository(self):
        self.assertNotEqual(
            ForceScheduler(
                name="testched", builderNames=[],
                codebases=oneCodebase(
                    repository=FixedParameter("repository", "git://pond.org/fisher.git"))),
            ForceScheduler(
                name="testched", builderNames=[],
                codebases=oneCodebase(
                    repository=FixedParameter("repository", "svn://ocean.com/trawler/"))))

    def test_compare_project(self):
        self.assertNotEqual(
            ForceScheduler(
                name="testched", builderNames=[],
                codebases=oneCodebase(
                    project=FixedParameter("project", "fisher"))),
            ForceScheduler(
                name="testched", builderNames=[],
                codebases=oneCodebase(
                    project=FixedParameter("project", "trawler"))))

    def test_compare_username(self):
        self.assertNotEqual(
            ForceScheduler(name="testched", builderNames=[]),
            ForceScheduler(name="testched", builderNames=[],
                           username=FixedParameter("username",
                                                   "The Fisher King <avallach@atlantis.al>")))

    def test_compare_properties(self):
        self.assertNotEqual(
            ForceScheduler(name="testched", builderNames=[],
                           properties=[]),
            ForceScheduler(name="testched", builderNames=[],
                           properties=[FixedParameter("prop", "thanks for the fish!")]))

    def test_compare_codebases(self):
        self.assertNotEqual(
            ForceScheduler(name="testched", builderNames=[],
                           codebases=['bar']),
            ForceScheduler(name="testched", builderNames=[],
                           codebases=['foo']))

    @defer.inlineCallbacks
    def test_basicForce(self):
        sched = self.makeScheduler()

        res = yield sched.force('user', builderNames=['a'], branch='a', reason='because',
                                revision='c', repository='d', project='p')

        # only one builder forced, so there should only be one brid
        self.assertEqual(res, (500, {1000: 100}))
        self.assertEqual(self.addBuildsetCalls, [
            ('addBuildsetForSourceStampsWithDefaults', dict(
                builderNames=['a'],
                waited_for=False,
                properties={
                    'owner': ('user', 'Force Build Form'),
                    'reason': ('because', 'Force Build Form'),
                },
                reason="A build was forced by 'user': because",
                sourcestamps=[
                    {'codebase': '', 'branch': 'a', 'revision': 'c',
                     'repository': 'd', 'project': 'p'},
                ])),
        ])

    @defer.inlineCallbacks
    def test_basicForce_reasonString(self):
        """Same as above, but with a reasonString"""
        sched = self.makeScheduler(
            reasonString='%(owner)s wants it %(reason)s')

        res = yield sched.force('user', builderNames=['a'], branch='a', reason='because',
                                revision='c', repository='d', project='p')
        _, brids = res

        # only one builder forced, so there should only be one brid
        self.assertEqual(len(brids), 1)

        self.assertEqual(self.addBuildsetCalls, [
            ('addBuildsetForSourceStampsWithDefaults', {
                'builderNames': ['a'],
                'properties': {'owner': ('user', 'Force Build Form'),
                               'reason': ('because', 'Force Build Form')},
                'reason': 'user wants it because',
                'sourcestamps': [{'branch': 'a',
                                  'codebase': '',
                                  'project': 'p',
                                  'repository': 'd',
                                  'revision': 'c'}],
                'waited_for': False}),
        ])

    @defer.inlineCallbacks
    def test_force_allBuilders(self):
        sched = self.makeScheduler()

        res = yield sched.force('user', branch='a', reason='because', revision='c',
                                repository='d', project='p',
                                )
        self.assertEqual(res, (500, {1000: 100, 1001: 101}))
        self.assertEqual(self.addBuildsetCalls, [
            ('addBuildsetForSourceStampsWithDefaults', dict(
                builderNames=['a', 'b'],
                waited_for=False,
                properties={
                    'owner': ('user', 'Force Build Form'),
                    'reason': ('because', 'Force Build Form'),
                },
                reason="A build was forced by 'user': because",
                sourcestamps=[
                    {'codebase': '', 'branch': 'a', 'revision': 'c',
                     'repository': 'd', 'project': 'p'},
                ])),
        ])

    @defer.inlineCallbacks
    def test_force_someBuilders(self):
        sched = self.makeScheduler(builderNames=['a', 'b', 'c'])

        res = yield sched.force('user', builderNames=['a', 'b'],
                                branch='a', reason='because', revision='c',
                                repository='d', project='p',
                                )
        self.assertEqual(res, (500, {1000: 100, 1001: 101}))
        self.assertEqual(self.addBuildsetCalls, [
            ('addBuildsetForSourceStampsWithDefaults', dict(
                builderNames=['a', 'b'],
                waited_for=False,
                properties={
                    'owner': ('user', 'Force Build Form'),
                    'reason': ('because', 'Force Build Form'),
                },
                reason="A build was forced by 'user': because",
                sourcestamps=[
                    {'codebase': '', 'branch': 'a', 'revision': 'c',
                     'repository': 'd', 'project': 'p'},
                ])),
        ])

    def test_bad_codebases(self):

        # codebases must be a list of either string or BaseParameter types
        with self.assertRaisesConfigError(
                "ForceScheduler 'foo': 'codebases' must be a "
                "list of strings or CodebaseParameter objects:"):
            ForceScheduler(name='foo', builderNames=['bar'], codebases=[123],)

        with self.assertRaisesConfigError(
                "ForceScheduler 'foo': 'codebases' must be a "
                "list of strings or CodebaseParameter objects:"):
            ForceScheduler(name='foo', builderNames=['bar'],
                           codebases=[IntParameter('foo')])

        # codebases cannot be empty
        with self.assertRaisesConfigError(
                "ForceScheduler 'foo': 'codebases' cannot be "
                "empty; use [CodebaseParameter(codebase='', hide=True)] if needed:"):
            ForceScheduler(name='foo', builderNames=['bar'], codebases=[])

        # codebases cannot be a dictionary
        # dictType on Python 3 is: "<class 'dict'>"
        # dictType on Python 2 is: "<type 'dict'>"
        dictType = str(type({}))
        errMsg = ("ForceScheduler 'foo': 'codebases' should be a list "
                  "of strings or CodebaseParameter, "
                  f"not {dictType}")
        with self.assertRaisesConfigError(errMsg):
            ForceScheduler(name='foo', builderNames=['bar'],
                           codebases={'cb': {'branch': 'trunk'}})

    @defer.inlineCallbacks
    def test_good_codebases(self):
        sched = self.makeScheduler(codebases=['foo', CodebaseParameter('bar')])
        yield sched.force('user', builderNames=['a'], reason='because',
                          foo_branch='a', foo_revision='c', foo_repository='d',
                          foo_project='p',
                          bar_branch='a2', bar_revision='c2', bar_repository='d2',
                          bar_project='p2'
                          )

        expProperties = {
            'owner': ('user', 'Force Build Form'),
            'reason': ('because', 'Force Build Form'),
        }
        self.assertEqual(self.addBuildsetCalls, [
            ('addBuildsetForSourceStampsWithDefaults', dict(
                builderNames=['a'],
                waited_for=False,
                properties=expProperties,
                reason="A build was forced by 'user': because",
                sourcestamps=[
                    {'branch': 'a2', 'project': 'p2', 'repository': 'd2',
                        'revision': 'c2', 'codebase': 'bar'},
                    {'branch': 'a', 'project': 'p', 'repository': 'd',
                        'revision': 'c', 'codebase': 'foo'},
                ])),
        ])

    @defer.inlineCallbacks
    def test_codebase_with_patch(self):
        sched = self.makeScheduler(codebases=['foo', CodebaseParameter('bar',
                                                                       patch=PatchParameter())])
        yield sched.force('user', builderNames=['a'], reason='because',
                          foo_branch='a', foo_revision='c', foo_repository='d',
                          foo_project='p',
                          bar_branch='a2', bar_revision='c2', bar_repository='d2',
                          bar_project='p2', bar_patch_body=b"xxx")
        expProperties = {
            'owner': ('user', 'Force Build Form'),
            'reason': ('because', 'Force Build Form'),
        }

        self.assertEqual(self.addBuildsetCalls, [
            ('addBuildsetForSourceStampsWithDefaults', dict(
                builderNames=['a'],
                waited_for=False,
                properties=expProperties,
                reason="A build was forced by 'user': because",
                sourcestamps=[
                    {'branch': 'a2', 'project': 'p2', 'repository': 'd2',
                        'revision': 'c2', 'codebase': 'bar',
                        'patch_body': b'xxx', 'patch_author': '', 'patch_subdir': '.',
                        'patch_comment': '', 'patch_level': 1},
                    {'branch': 'a', 'project': 'p', 'repository': 'd',
                        'revision': 'c', 'codebase': 'foo'},
                ])),
        ])

    def formatJsonForTest(self, gotJson):
        ret = ""
        linestart = "expectJson='"
        spaces = 7 * 4 + 2
        while len(gotJson) > (90 - spaces):
            gotJson = " " * spaces + linestart + gotJson
            pos = gotJson[:100].rfind(",")
            if pos > 0:
                pos += 2
            ret += gotJson[:pos] + "'\n"
            gotJson = gotJson[pos:]
            linestart = "'"
        ret += " " * spaces + linestart + gotJson + "')\n"
        return ret

    # value = the value to be sent with the parameter (ignored if req is set)
    # expect = the expected result (can be an exception type)
    # klass = the parameter class type
    # req = use this request instead of the auto-generated one based on value
    @defer.inlineCallbacks
    def do_ParameterTest(self,
                         expect,
                         klass,
                         # None=one prop, Exception=exception, dict=many props
                         expectKind=None,
                         owner='user',
                         value=None, req=None,
                         expectJson=None,
                         **kwargs):

        name = kwargs.setdefault('name', 'p1')

        # construct one if needed
        if isinstance(klass, type):
            prop = klass(**kwargs)
        else:
            prop = klass

        self.assertEqual(prop.name, name)
        self.assertEqual(prop.label, kwargs.get('label', prop.name))
        if expectJson is not None:
            gotSpec = prop.getSpec()
            gotJson = json.dumps(gotSpec)
            expectSpec = json.loads(expectJson)
            if gotSpec != expectSpec:
                try:
                    import xerox  # pylint: disable=import-outside-toplevel
                    formatted = self.formatJsonForTest(gotJson)
                    print(
                        "You may update the test with (copied to clipboard):\n" + formatted)
                    xerox.copy(formatted)
                    input()
                except ImportError:
                    print("Note: for quick fix, pip install xerox")
            self.assertEqual(gotSpec, expectSpec)

        sched = self.makeScheduler(properties=[prop])

        if not req:
            req = {name: value, 'reason': 'because'}
        try:
            bsid, brids = yield sched.force(owner, builderNames=['a'], **req)
        except Exception as e:
            if expectKind is not Exception:
                # an exception is not expected
                raise
            if not isinstance(e, expect):
                # the exception is the wrong kind
                raise
            return None  # success

        expect_props = {
            'owner': ('user', 'Force Build Form'),
            'reason': ('because', 'Force Build Form'),
        }

        if expectKind is None:
            expect_props[name] = (expect, 'Force Build Form')
        elif expectKind is dict:
            for k, v in expect.items():
                expect_props[k] = (v, 'Force Build Form')
        else:
            self.fail("expectKind is wrong type!")

        # only forced on 'a'
        self.assertEqual((bsid, brids), (500, {1000: 100}))
        self.assertEqual(self.addBuildsetCalls, [
            ('addBuildsetForSourceStampsWithDefaults', dict(
                builderNames=['a'],
                waited_for=False,
                properties=expect_props,
                reason="A build was forced by 'user': because",
                sourcestamps=[
                    {'branch': '', 'project': '', 'repository': '',
                     'revision': '', 'codebase': ''},
                ])),
        ])
        return None

    def test_StringParameter(self):
        self.do_ParameterTest(value="testedvalue", expect="testedvalue",
                              klass=StringParameter,
                              expectJson='{"name": "p1", "fullName": "p1", "label": "p1", '
                              '"tablabel": "p1", "type": "text", "default": "", "required": false, '
                              '"multiple": false, "regex": null, "hide": false, "maxsize": null, '
                              '"size": 10, "autopopulate": null}')

    def test_StringParameter_Required(self):
        self.do_ParameterTest(value=" ", expect=CollectedValidationError,
                              expectKind=Exception,
                              klass=StringParameter, required=True)

    def test_StringParameter_maxsize(self):
        self.do_ParameterTest(value="xx" * 20, expect=CollectedValidationError,
                              expectKind=Exception,
                              klass=StringParameter, maxsize=10)

    def test_FileParameter_maxsize(self):
        self.do_ParameterTest(value="xx" * 20, expect=CollectedValidationError,
                              expectKind=Exception,
                              klass=FileParameter, maxsize=10)

    def test_FileParameter(self):
        self.do_ParameterTest(value="xx", expect="xx",
                              klass=FileParameter,
                              expectJson='{"name": "p1", "fullName": "p1", "label": "p1", '
                              '"tablabel": "p1", "type": "file", "default": "", "required": false, '
                              '"multiple": false, "regex": null, "hide": false, '
                              '"maxsize": 10485760, "autopopulate": null}')

    def test_PatchParameter(self):
        expect_json = (
            '{"name": "p1", "fullName": "p1", "label": "p1", "autopopulate": null, '
            '"tablabel": "p1", "type": "nested", "default": "", "required": false, '
            '"multiple": false, "regex": null, "hide": false, "maxsize": null, '
            '"layout": "vertical", "columns": 1, "fields": [{"name": "body", '
            '"fullName": "p1_body", "label": "body", "tablabel": "body", "autopopulate": null, '
            '"type": "file", "default": "", "required": false, "multiple": false, '
            '"regex": null, "hide": false, "maxsize": 10485760}, {"name": "level", '
            '"fullName": "p1_level", "label": "level", "tablabel": "level", '
            '"type": "int", "default": 1, "required": false, "multiple": false, '
            '"regex": null, "hide": false, "maxsize": null, "size": 10, "autopopulate": null}, '
            '{"name": "author", "fullName": "p1_author", "label": "author", '
            '"tablabel": "author", "type": "text", "default": "", "autopopulate": null, '
            '"required": false, "multiple": false, "regex": null, "hide": false, '
            '"maxsize": null, "size": 10}, {"name": "comment", "autopopulate": null, '
            '"fullName": "p1_comment", "label": "comment", "tablabel": "comment", '
            '"type": "text", "default": "", "required": false, "multiple": false, '
            '"regex": null, "hide": false, "maxsize": null, "size": 10}, '
            '{"name": "subdir", "fullName": "p1_subdir", "label": "subdir", '
            '"tablabel": "subdir", "type": "text", "default": ".", "autopopulate": null, '
            '"required": false, "multiple": false, "regex": null, "hide": false, '
            '"maxsize": null, "size": 10}]}'
        )

        self.do_ParameterTest(req=dict(p1_author='me', reason="because"), expect={
                                   'author': 'me',
                                   'body': '',
                                   'comment': '',
                                   'level': 1,
                                   'subdir': '.'},
                              klass=PatchParameter,
                              expectJson=expect_json)

    def test_IntParameter(self):
        self.do_ParameterTest(value="123", expect=123, klass=IntParameter,
                              expectJson='{"name": "p1", "fullName": "p1", "label": "p1", '
                              '"tablabel": "p1", "type": "int", "default": 0, "required": false, '
                              '"multiple": false, "regex": null, "hide": false, "maxsize": null, '
                              '"size": 10, "autopopulate": null}')

    def test_FixedParameter(self):
        self.do_ParameterTest(value="123", expect="321", klass=FixedParameter,
                              default="321",
                              expectJson='{"name": "p1", "fullName": "p1", "label": "p1", '
                              '"tablabel": "p1", "type": "fixed", "default": "321", '
                              '"required": false, "multiple": false, "regex": null, "hide": true, '
                              '"maxsize": null, "autopopulate": null}')

    def test_BooleanParameter_True(self):
        req = dict(p1=True, reason='because')
        self.do_ParameterTest(value="123", expect=True, klass=BooleanParameter,
                              req=req,
                              expectJson='{"name": "p1", "fullName": "p1", "label": "p1", '
                              '"tablabel": "p1", "type": "bool", "default": "", "required": false, '
                              '"multiple": false, "regex": null, "hide": false, '
                              '"maxsize": null, "autopopulate": null}')

    def test_BooleanParameter_False(self):
        req = dict(p2=True, reason='because')
        self.do_ParameterTest(value="123", expect=False,
                              klass=BooleanParameter, req=req)

    def test_UserNameParameter(self):
        email = "test <test@buildbot.net>"
        expect_json = (
            '{"name": "username", "fullName": "username", '
            '"label": "Your name:", "tablabel": "Your name:", "type": "username", '
            '"default": "", "required": false, "multiple": false, "regex": null, '
            '"hide": false, "maxsize": null, "size": 30, '
            '"need_email": true, "autopopulate": null}'
        )
        self.do_ParameterTest(value=email, expect=email,
                              klass=UserNameParameter(),
                              name="username", label="Your name:",
                              expectJson=expect_json)

    def test_UserNameParameterIsValidMail(self):
        email = "test@buildbot.net"
        expect_json = (
            '{"name": "username", "fullName": "username", '
            '"label": "Your name:", "tablabel": "Your name:", "type": "username", '
            '"default": "", "required": false, "multiple": false, "regex": null, '
            '"hide": false, "maxsize": null, "size": 30, '
            '"need_email": true, "autopopulate": null}'
        )
        self.do_ParameterTest(value=email, expect=email,
                              klass=UserNameParameter(),
                              name="username", label="Your name:",
                              expectJson=expect_json)

    def test_UserNameParameterIsValidMailBis(self):
        email = "<test@buildbot.net>"
        expect_json = (
            '{"name": "username", "fullName": "username", '
            '"label": "Your name:", "tablabel": "Your name:", "type": "username", '
            '"default": "", "required": false, "multiple": false, "regex": null, '
            '"hide": false, "maxsize": null, "size": 30, '
            '"need_email": true, "autopopulate": null}'
        )
        self.do_ParameterTest(value=email, expect=email,
                              klass=UserNameParameter(),
                              name="username", label="Your name:",
                              expectJson=expect_json)

    def test_ChoiceParameter(self):
        self.do_ParameterTest(value='t1', expect='t1',
                              klass=ChoiceStringParameter, choices=[
                                  't1', 't2'],
                              expectJson='{"name": "p1", "fullName": "p1", "label": "p1", '
                              '"tablabel": "p1", "type": "list", "default": "", "required": false, '
                              '"multiple": false, "regex": null, "hide": false, "maxsize": null, '
                              '"choices": ["t1", "t2"], "strict": true, "autopopulate": null}')

    def test_ChoiceParameterError(self):
        self.do_ParameterTest(value='t3',
                              expect=CollectedValidationError,
                              expectKind=Exception,
                              klass=ChoiceStringParameter, choices=[
                                  't1', 't2'],
                              debug=False)

    def test_ChoiceParameterError_notStrict(self):
        self.do_ParameterTest(value='t1', expect='t1',
                              strict=False,
                              klass=ChoiceStringParameter, choices=['t1', 't2'])

    def test_ChoiceParameterMultiple(self):
        self.do_ParameterTest(value=['t1', 't2'], expect=['t1', 't2'],
                              klass=ChoiceStringParameter, choices=['t1', 't2'], multiple=True,
                              expectJson='{"name": "p1", "fullName": "p1", "label": "p1", '
                              '"tablabel": "p1", "type": "list", "default": "", "required": false, '
                              '"multiple": true, "regex": null, "hide": false, "maxsize": null, '
                              '"choices": ["t1", "t2"], "strict": true, "autopopulate": null}')

    def test_ChoiceParameterMultipleError(self):
        self.do_ParameterTest(value=['t1', 't3'],
                              expect=CollectedValidationError,
                              expectKind=Exception,
                              klass=ChoiceStringParameter, choices=[
                                  't1', 't2'],
                              multiple=True, debug=False)

    def test_NestedParameter(self):
        fields = [
            IntParameter(name="foo")
        ]
        expect_json = (
            '{"name": "p1", "fullName": "p1", "label": "p1", "autopopulate": null, '
            '"tablabel": "p1", "type": "nested", "default": "", "required": false, '
            '"multiple": false, "regex": null, "hide": false, "maxsize": null, '
            '"layout": "vertical", "columns": 1, "fields": [{"name": "foo", '
            '"fullName": "p1_foo", "label": "foo", "tablabel": "foo", "autopopulate": null, '
            '"type": "int", "default": 0, "required": false, "multiple": false, '
            '"regex": null, "hide": false, "maxsize": null, "size": 10}]}'
        )
        self.do_ParameterTest(req=dict(p1_foo='123', reason="because"),
                              expect=dict(foo=123),
                              klass=NestedParameter, fields=fields,
                              expectJson=expect_json)

    def test_NestedNestedParameter(self):
        fields = [
            NestedParameter(name="inner", fields=[
                StringParameter(name='str'),
                AnyPropertyParameter(name='any')
            ]),
            IntParameter(name="foo")
        ]
        self.do_ParameterTest(req=dict(p1_foo='123',
                                       p1_inner_str="bar",
                                       p1_inner_any_name="hello",
                                       p1_inner_any_value="world",
                                       reason="because"),
                              expect=dict(
                                  foo=123, inner=dict(str="bar", hello="world")),
                              klass=NestedParameter, fields=fields)

    def test_NestedParameter_nullname(self):
        # same as above except "p1" and "any" are skipped
        fields = [
            NestedParameter(name="inner", fields=[
                StringParameter(name='str'),
                AnyPropertyParameter(name='')
            ]),
            IntParameter(name="foo"),
            NestedParameter(name='bar', fields=[
                NestedParameter(
                    name='', fields=[AnyPropertyParameter(name='a')]),
                NestedParameter(
                    name='', fields=[AnyPropertyParameter(name='b')])
            ])
        ]
        self.do_ParameterTest(req=dict(foo='123',
                                       inner_str="bar",
                                       inner_name="hello",
                                       inner_value="world",
                                       reason="because",
                                       bar_a_name="a",
                                       bar_a_value="7",
                                       bar_b_name="b",
                                       bar_b_value="8"),
                              expect=dict(foo=123,
                                          inner=dict(str="bar", hello="world"),
                                          bar={'a': '7', 'b': '8'}),
                              expectKind=dict,
                              klass=NestedParameter, fields=fields, name='')

    def test_bad_reason(self):
        with self.assertRaisesConfigError(
                "ForceScheduler 'testsched': reason must be a StringParameter"):
            ForceScheduler(name='testsched', builderNames=[],
                           codebases=['bar'], reason="foo")

    def test_bad_username(self):
        with self.assertRaisesConfigError(
                "ForceScheduler 'testsched': username must be a StringParameter"):
            ForceScheduler(name='testsched', builderNames=[],
                           codebases=['bar'], username="foo")

    def test_notstring_name(self):
        with self.assertRaisesConfigError(
                "ForceScheduler name must be a unicode string:"):
            ForceScheduler(name=1234, builderNames=[], codebases=['bar'],
                           username="foo")

    def test_notidentifier_name(self):
        # FIXME: this test should be removed eventually when bug 3460 gets a
        # real fix
        with self.assertRaisesConfigError(
                "ForceScheduler name must be an identifier: 'my scheduler'"):
            ForceScheduler(name='my scheduler', builderNames=[],
                           codebases=['bar'], username="foo")

    def test_emptystring_name(self):
        with self.assertRaisesConfigError(
                "ForceScheduler name must not be empty:"):
            ForceScheduler(name='', builderNames=[], codebases=['bar'],
                           username="foo")

    def test_integer_builderNames(self):
        with self.assertRaisesConfigError(
                "ForceScheduler 'testsched': builderNames must be a list of strings:"):
            ForceScheduler(name='testsched', builderNames=1234,
                           codebases=['bar'], username="foo")

    def test_listofints_builderNames(self):
        with self.assertRaisesConfigError(
                "ForceScheduler 'testsched': builderNames must be a list of strings:"):
            ForceScheduler(name='testsched', builderNames=[1234],
                           codebases=['bar'], username="foo")

    def test_listofunicode_builderNames(self):
        ForceScheduler(name='testsched', builderNames=['a', 'b'])

    def test_listofmixed_builderNames(self):
        with self.assertRaisesConfigError(
                "ForceScheduler 'testsched': builderNames must be a list of strings:"):
            ForceScheduler(name='testsched', builderNames=['test', 1234],
                           codebases=['bar'], username="foo")

    def test_integer_properties(self):
        with self.assertRaisesConfigError(
                "ForceScheduler 'testsched': properties must be a list of BaseParameters:"):
            ForceScheduler(name='testsched', builderNames=[],
                           codebases=['bar'], username="foo",
                           properties=1234)

    def test_listofints_properties(self):
        with self.assertRaisesConfigError(
                "ForceScheduler 'testsched': properties must be a list of BaseParameters:"):
            ForceScheduler(name='testsched', builderNames=[],
                           codebases=['bar'], username="foo",
                           properties=[1234, 2345])

    def test_listofmixed_properties(self):
        with self.assertRaisesConfigError(
                "ForceScheduler 'testsched': properties must be a list of BaseParameters:"):
            ForceScheduler(name='testsched', builderNames=[],
                           codebases=['bar'], username="foo",
                           properties=[BaseParameter(name="test",),
                           4567])

    def test_novalue_to_parameter(self):
        with self.assertRaisesConfigError(
                "Use default='1234' instead of value=... to give a default Parameter value"):
            BaseParameter(name="test", value="1234")



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
"""
HVAC based providers
"""

from twisted.internet import defer
from twisted.internet import threads

from buildbot import config
from buildbot.secrets.providers.base import SecretProviderBase


class VaultAuthenticator:
    """
    base HVAC authenticator class
    """

    def authenticate(self, client):
        pass


class VaultAuthenticatorToken(VaultAuthenticator):
    """
    HVAC authenticator for static token
    """

    def __init__(self, token):
        self.token = token

    def authenticate(self, client):
        client.token = self.token


class VaultAuthenticatorApprole(VaultAuthenticator):
    """
    HVAC authenticator for Approle login method
    """

    def __init__(self, roleId, secretId):
        self.roleId = roleId
        self.secretId = secretId

    def authenticate(self, client):
        client.auth.approle.login(role_id=self.roleId, secret_id=self.secretId)


class HashiCorpVaultKvSecretProvider(SecretProviderBase):
    """
    Basic provider where each secret is stored in Vault KV secret engine.
    In case more secret engines are going to be supported, each engine should have it's own class.
    """

    name = 'SecretInVaultKv'

    def checkConfig(self, vault_server=None, authenticator=None, secrets_mount=None,
                    api_version=2, path_delimiter='|', path_escape='\\'):
        try:
            import hvac
            [hvac]
        except ImportError:  # pragma: no cover
            config.error(f"{self.__class__.__name__} needs the hvac package installed " +
                         "(pip install hvac)")

        if not isinstance(vault_server, str):
            config.error(f"vault_server must be a string while it is {type(vault_server)}")
        if not isinstance(path_delimiter, str) or len(path_delimiter) > 1:
            config.error("path_delimiter must be a single character")
        if not isinstance(path_escape, str) or len(path_escape) > 1:
            config.error("path_escape must be a single character")
        if not isinstance(authenticator, VaultAuthenticator):
            config.error("authenticator must be instance of VaultAuthenticator while it is "
                         f"{type(authenticator)}")

        if api_version not in [1, 2]:
            config.error(f"api_version {api_version} is not supported")

    def reconfigService(self, vault_server=None, authenticator=None, secrets_mount=None,
                        api_version=2, path_delimiter='|', path_escape='\\'):
        try:
            import hvac
        except ImportError:  # pragma: no cover
            config.error(f"{self.__class__.__name__} needs the hvac package installed " +
                         "(pip install hvac)")

        if secrets_mount is None:
            secrets_mount = "secret"
        self.secrets_mount = secrets_mount
        self.path_delimiter = path_delimiter
        self.path_escape = path_escape
        self.authenticator = authenticator
        self.api_version = api_version
        if vault_server.endswith('/'):  # pragma: no cover
            vault_server = vault_server[:-1]
        self.client = hvac.Client(vault_server)
        self.client.secrets.kv.default_kv_version = api_version
        return self

    def escaped_split(self, s):
        """
        parse and split string, respecting escape characters
        """
        ret = []
        current = []
        itr = iter(s)
        for ch in itr:
            if ch == self.path_escape:
                try:
                    # skip the next character; it has been escaped and remove
                    # escape character
                    current.append(next(itr))
                except StopIteration:
                    # escape character on end of the string is safest to ignore, as buildbot for
                    # each secret identifier tries all secret providers until value is found,
                    # meaning we may end up parsing identifiers for other secret providers, where
                    # our escape character may be valid on end of string
                    pass
            elif ch == self.path_delimiter:
                # split! (add current to the list and reset it)
                ret.append(''.join(current))
                current = []
            else:
                current.append(ch)
        ret.append(''.join(current))
        return ret

    def thd_hvac_wrap_read(self, path):
        if self.api_version == 1:
            return self.client.secrets.kv.v1.read_secret(path=path, mount_point=self.secrets_mount)
        else:
            return self.client.secrets.kv.v2.read_secret_version(path=path,
                                                                 mount_point=self.secrets_mount)

    def thd_hvac_get(self, path):
        """
        query secret from Vault and re-authenticate if not authenticated
        """

        if not self.client.is_authenticated():
            self.authenticator.authenticate(self.client)

        response = self.thd_hvac_wrap_read(path=path)

        return response

    @defer.inlineCallbacks
    def get(self, entry):
        """
        get the value from vault secret backend
        """

        parts = self.escaped_split(entry)
        if len(parts) == 1:
            raise KeyError("Vault secret specification must contain attribute name separated from "
                           f"path by '{self.path_delimiter}'")
        if len(parts) > 2:
            raise KeyError(f"Multiple separators ('{self.path_delimiter}') found in vault "
                           f"path '{entry}'. All occurences of '{self.path_delimiter}' in path or "
                           f"attribute name must be escaped using '{self.path_escape}'")

        name = parts[0]
        key = parts[1]

        response = yield threads.deferToThread(self.thd_hvac_get, path=name)

        # in KVv2 we have extra "data" dictionary, as vault provides metadata as well
        if self.api_version == 2:
            response = response['data']

        try:
            return response['data'][key]
        except KeyError as e:
            raise KeyError(
                f"The secret {entry} does not exist in Vault provider: {e}") from e



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer

from buildbot.process.properties import Interpolate
from buildbot.test.fake.secrets import FakeSecretStorage
from buildbot.test.util.integration import RunMasterBase


class SecretsConfig(RunMasterBase):

    @defer.inlineCallbacks
    def test_secret(self):
        yield self.setupConfig(masterConfig())
        build = yield self.doForceBuild(wantSteps=True, wantLogs=True)
        self.assertEqual(build['buildid'], 1)
        res = yield self.checkBuildStepLogExist(build, "<foo>")
        self.assertTrue(res)

    @defer.inlineCallbacks
    def test_withsecrets(self):
        yield self.setupConfig(masterConfig(use_with=True))
        build = yield self.doForceBuild(wantSteps=True, wantLogs=True)
        self.assertEqual(build['buildid'], 1)
        res = yield self.checkBuildStepLogExist(build, "<foo>")
        self.assertTrue(res)


# master configuration
def masterConfig(use_with=False):
    c = {}
    from buildbot.config import BuilderConfig
    from buildbot.process.factory import BuildFactory
    from buildbot.plugins import schedulers, steps

    c['schedulers'] = [
        schedulers.ForceScheduler(
            name="force",
            builderNames=["testy"])]

    c['secretsProviders'] = [FakeSecretStorage(
        secretdict={"foo": "bar", "something": "more"})]
    f = BuildFactory()
    if use_with:
        secrets_list = [("pathA", Interpolate('%(secret:something)s'))]
        with f.withSecrets(secrets_list):
            f.addStep(steps.ShellCommand(command=Interpolate('echo %(secret:foo)s')))
    else:
        f.addSteps([steps.ShellCommand(command=Interpolate('echo %(secret:foo)s'))],
                   withSecrets=[("pathA", Interpolate('%(secret:something)s'))])
    c['builders'] = [
        BuilderConfig(name="testy",
                      workernames=["local1"],
                      factory=f)]
    return c



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import mock

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.data import builds
from buildbot.data import resultspec
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import endpoint
from buildbot.test.util import interfaces
from buildbot.util import epoch2datetime


class BuildEndpoint(endpoint.EndpointMixin, unittest.TestCase):

    endpointClass = builds.BuildEndpoint
    resourceTypeClass = builds.Build

    def setUp(self):
        self.setUpEndpoint()
        self.db.insertTestData([
            fakedb.Builder(id=77, name='builder77'),
            fakedb.Master(id=88),
            fakedb.Worker(id=13, name='wrk'),
            fakedb.Buildset(id=8822),
            fakedb.BuildRequest(id=82, buildsetid=8822, builderid=77),
            fakedb.Build(id=13, builderid=77, masterid=88, workerid=13,
                         buildrequestid=82, number=3),
            fakedb.Build(id=14, builderid=77, masterid=88, workerid=13,
                         buildrequestid=82, number=4),
            fakedb.Build(id=15, builderid=77, masterid=88, workerid=13,
                         buildrequestid=82, number=5),
            fakedb.BuildProperty(buildid=13, name='reason', value='"force build"',
                                 source="Force Build Form"),
        ])

    def tearDown(self):
        self.tearDownEndpoint()

    @defer.inlineCallbacks
    def test_get_existing(self):
        build = yield self.callGet(('builds', 14))
        self.validateData(build)
        self.assertEqual(build['number'], 4)

    @defer.inlineCallbacks
    def test_get_missing(self):
        build = yield self.callGet(('builds', 9999))
        self.assertEqual(build, None)

    @defer.inlineCallbacks
    def test_get_missing_builder_number(self):
        build = yield self.callGet(('builders', 999, 'builds', 4))
        self.assertEqual(build, None)

    @defer.inlineCallbacks
    def test_get_builder_missing_number(self):
        build = yield self.callGet(('builders', 77, 'builds', 44))
        self.assertEqual(build, None)

    @defer.inlineCallbacks
    def test_get_builder_number(self):
        build = yield self.callGet(('builders', 77, 'builds', 5))
        self.validateData(build)
        self.assertEqual(build['buildid'], 15)

    @defer.inlineCallbacks
    def test_get_buildername_number(self):
        build = yield self.callGet(('builders', 'builder77', 'builds', 5))
        self.validateData(build)
        self.assertEqual(build['buildid'], 15)

    @defer.inlineCallbacks
    def test_get_buildername_not_existing_number(self):
        build = yield self.callGet(('builders', 'builder77_nope', 'builds', 5))
        self.assertEqual(build, None)

    @defer.inlineCallbacks
    def test_properties_injection(self):
        resultSpec = resultspec.OptimisedResultSpec(
            properties=[resultspec.Property(b'property', 'eq', 'reason')])
        build = yield self.callGet(('builders', 77, 'builds', 3), resultSpec=resultSpec)
        self.validateData(build)
        self.assertIn('reason', build['properties'])

    @defer.inlineCallbacks
    def test_action_stop(self):
        yield self.callControl("stop", {}, ('builders', 77, 'builds', 5))
        self.master.mq.assertProductions(
            [(('control', 'builds', '15', 'stop'), {'reason': 'no reason'})])

    @defer.inlineCallbacks
    def test_action_stop_reason(self):
        yield self.callControl("stop", {'reason': 'because'}, ('builders', 77, 'builds', 5))
        self.master.mq.assertProductions(
            [(('control', 'builds', '15', 'stop'), {'reason': 'because'})])

    @defer.inlineCallbacks
    def test_action_rebuild(self):
        self.patch(self.master.data.updates, "rebuildBuildrequest",
                   mock.Mock(spec=self.master.data.updates.rebuildBuildrequest,
                             return_value=(1, [2])))
        r = yield self.callControl("rebuild", {}, ('builders', 77, 'builds', 5))
        self.assertEqual(r, (1, [2]))

        buildrequest = yield self.master.data.get(('buildrequests', 82))
        self.master.data.updates.rebuildBuildrequest.assert_called_with(
            buildrequest)


class BuildsEndpoint(endpoint.EndpointMixin, unittest.TestCase):

    endpointClass = builds.BuildsEndpoint
    resourceTypeClass = builds.Build

    def setUp(self):
        self.setUpEndpoint()
        self.db.insertTestData([
            fakedb.Builder(id=77, name='builder77'),
            fakedb.Builder(id=78, name='builder78'),
            fakedb.Builder(id=79, name='builder79'),
            fakedb.Master(id=88),
            fakedb.Worker(id=13, name='wrk'),
            fakedb.Buildset(id=8822),
            fakedb.BuildRequest(id=82, buildsetid=8822),
            fakedb.Build(id=13, builderid=77, masterid=88, workerid=13,
                         buildrequestid=82, number=3),
            fakedb.Build(id=14, builderid=77, masterid=88, workerid=13,
                         buildrequestid=82, number=4),
            fakedb.Build(id=15, builderid=78, masterid=88, workerid=12,
                         buildrequestid=83, number=5, complete_at=1),
            fakedb.Build(id=16, builderid=79, masterid=88, workerid=12,
                         buildrequestid=84, number=6, complete_at=1),
            fakedb.BuildProperty(buildid=13, name='reason', value='"force build"',
                                 source="Force Build Form"),
        ])

    def tearDown(self):
        self.tearDownEndpoint()

    @defer.inlineCallbacks
    def test_get_all(self):
        builds = yield self.callGet(('builds',))

        for build in builds:
            self.validateData(build)

        self.assertEqual(sorted([b['number'] for b in builds]),
                         [3, 4, 5, 6])

    @defer.inlineCallbacks
    def test_get_builder(self):
        builds = yield self.callGet(('builders', 78, 'builds'))

        for build in builds:
            self.validateData(build)

        self.assertEqual(sorted([b['number'] for b in builds]), [5])

    @defer.inlineCallbacks
    def test_get_buildername(self):
        builds = yield self.callGet(('builders', 'builder78', 'builds'))

        for build in builds:
            self.validateData(build)

        self.assertEqual(sorted([b['number'] for b in builds]), [5])

    @defer.inlineCallbacks
    def test_get_buildername_not_existing(self):
        builds = yield self.callGet(('builders', 'builder78_nope', 'builds'))
        self.assertEqual(builds, [])

    @defer.inlineCallbacks
    def test_get_buildrequest(self):
        builds = yield self.callGet(('buildrequests', 82, 'builds'))

        for build in builds:
            self.validateData(build)

        self.assertEqual(sorted([b['number'] for b in builds]), [3, 4])

    @defer.inlineCallbacks
    def test_get_buildrequest_not_existing(self):
        builds = yield self.callGet(('buildrequests', 899, 'builds'))
        self.assertEqual(builds, [])

    @defer.inlineCallbacks
    def test_get_buildrequest_via_filter(self):
        resultSpec = resultspec.OptimisedResultSpec(
            filters=[resultspec.Filter('buildrequestid', 'eq', [82])])
        builds = yield self.callGet(('builds',), resultSpec=resultSpec)

        for build in builds:
            self.validateData(build)

        self.assertEqual(sorted([b['number'] for b in builds]), [3, 4])

    @defer.inlineCallbacks
    def test_get_buildrequest_via_filter_with_string(self):
        resultSpec = resultspec.OptimisedResultSpec(
            filters=[resultspec.Filter('buildrequestid', 'eq', ['82'])])
        builds = yield self.callGet(('builds',), resultSpec=resultSpec)

        for build in builds:
            self.validateData(build)

        self.assertEqual(sorted([b['number'] for b in builds]), [3, 4])

    @defer.inlineCallbacks
    def test_get_worker(self):
        builds = yield self.callGet(('workers', 13, 'builds'))

        for build in builds:
            self.validateData(build)

        self.assertEqual(sorted([b['number'] for b in builds]), [3, 4])

    @defer.inlineCallbacks
    def test_get_complete(self):
        resultSpec = resultspec.OptimisedResultSpec(
            filters=[resultspec.Filter('complete', 'eq', [False])])
        builds = yield self.callGet(('builds',), resultSpec=resultSpec)

        for build in builds:
            self.validateData(build)

        self.assertEqual(sorted([b['number'] for b in builds]), [3, 4])

    @defer.inlineCallbacks
    def test_get_complete_at(self):
        resultSpec = resultspec.OptimisedResultSpec(
            filters=[resultspec.Filter('complete_at', 'eq', [None])])
        builds = yield self.callGet(('builds',), resultSpec=resultSpec)

        for build in builds:
            self.validateData(build)

        self.assertEqual(sorted([b['number'] for b in builds]), [3, 4])

    @defer.inlineCallbacks
    def test_properties_injection(self):
        resultSpec = resultspec.OptimisedResultSpec(
            properties=[resultspec.Property(b'property', 'eq', 'reason')])
        builds = yield self.callGet(('builds',), resultSpec=resultSpec)

        for build in builds:
            self.validateData(build)

        self.assertTrue(any(('reason' in b['properties']) for b in builds))

    @defer.inlineCallbacks
    def test_get_filter_eq(self):
        resultSpec = resultspec.OptimisedResultSpec(
            filters=[resultspec.Filter('builderid', 'eq', [78, 79])])
        builds = yield self.callGet(('builds',), resultSpec=resultSpec)

        for b in builds:
            self.validateData(b)

        self.assertEqual(sorted([b['number'] for b in builds]), [5, 6])

    @defer.inlineCallbacks
    def test_get_filter_ne(self):
        resultSpec = resultspec.OptimisedResultSpec(
            filters=[resultspec.Filter('builderid', 'ne', [78, 79])])
        builds = yield self.callGet(('builds',), resultSpec=resultSpec)

        for b in builds:
            self.validateData(b)

        self.assertEqual(sorted([b['number'] for b in builds]), [3, 4])


class Build(interfaces.InterfaceTests, TestReactorMixin, unittest.TestCase):
    new_build_event = {'builderid': 10,
                       'buildid': 100,
                       'buildrequestid': 13,
                       'workerid': 20,
                       'complete': False,
                       'complete_at': None,
                       'masterid': 824,
                       'number': 1,
                       'results': None,
                       'started_at': epoch2datetime(1),
                       'state_string': 'created',
                       'properties': {}}

    def setUp(self):
        self.setup_test_reactor()
        self.master = fakemaster.make_master(self, wantMq=True, wantDb=True,
                                             wantData=True)
        self.rtype = builds.Build(self.master)

    @defer.inlineCallbacks
    def do_test_callthrough(self, dbMethodName, method, exp_args=None,
                            exp_kwargs=None, *args, **kwargs):
        rv = (1, 2)
        m = mock.Mock(return_value=defer.succeed(rv))
        setattr(self.master.db.builds, dbMethodName, m)
        res = yield method(*args, **kwargs)
        self.assertIdentical(res, rv)
        m.assert_called_with(*(exp_args or args), **(exp_kwargs or kwargs))

    @defer.inlineCallbacks
    def do_test_event(self, method, exp_events=None,
                      *args, **kwargs):
        self.reactor.advance(1)
        if exp_events is None:
            exp_events = []
        yield method(*args, **kwargs)
        self.master.mq.assertProductions(exp_events)

    def test_signature_addBuild(self):
        @self.assertArgSpecMatches(
            self.master.data.updates.addBuild,  # fake
            self.rtype.addBuild)  # real
        def addBuild(self, builderid, buildrequestid, workerid):
            pass

    def test_addBuild(self):
        return self.do_test_callthrough('addBuild', self.rtype.addBuild,
                                        builderid=10, buildrequestid=13, workerid=20,
                                        exp_kwargs=dict(builderid=10, buildrequestid=13,
                                                        workerid=20, masterid=self.master.masterid,
                                                        state_string='created'))

    def test_addBuildEvent(self):

        @defer.inlineCallbacks
        def addBuild(*args, **kwargs):
            buildid, _ = yield self.rtype.addBuild(*args, **kwargs)
            yield self.rtype.generateNewBuildEvent(buildid)
            return None

        return self.do_test_event(addBuild,
                                  builderid=10, buildrequestid=13, workerid=20,
                                  exp_events=[(('builders', '10', 'builds', '1', 'new'),
                                               self.new_build_event),
                                              (('builds', '100', 'new'),
                                               self.new_build_event),
                                              (('workers', '20', 'builds', '100', 'new'),
                                               self.new_build_event)])

    def test_signature_setBuildStateString(self):
        @self.assertArgSpecMatches(
            self.master.data.updates.setBuildStateString,  # fake
            self.rtype.setBuildStateString)  # real
        def setBuildStateString(self, buildid, state_string):
            pass

    def test_setBuildStateString(self):
        return self.do_test_callthrough('setBuildStateString',
                                        self.rtype.setBuildStateString,
                                        buildid=10, state_string='a b')

    def test_signature_finishBuild(self):
        @self.assertArgSpecMatches(
            self.master.data.updates.finishBuild,  # fake
            self.rtype.finishBuild)  # real
        def finishBuild(self, buildid, results):
            pass

    def test_finishBuild(self):
        return self.do_test_callthrough('finishBuild', self.rtype.finishBuild,
                                        buildid=15, results=3)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import os

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.test.util import dirs
from buildbot.util import maildir


class TestMaildirService(dirs.DirsMixin, unittest.TestCase):

    def setUp(self):
        self.maildir = os.path.abspath("maildir")
        self.newdir = os.path.join(self.maildir, "new")
        self.curdir = os.path.join(self.maildir, "cur")
        self.tmpdir = os.path.join(self.maildir, "tmp")
        self.setUpDirs(self.maildir, self.newdir, self.curdir, self.tmpdir)

        self.svc = None

    def tearDown(self):
        if self.svc and self.svc.running:
            self.svc.stopService()
        self.tearDownDirs()

    # tests

    @defer.inlineCallbacks
    def test_start_stop_repeatedly(self):
        self.svc = maildir.MaildirService(self.maildir)
        self.svc.startService()
        yield self.svc.stopService()
        self.svc.startService()
        yield self.svc.stopService()
        self.assertEqual(len(list(self.svc)), 0)

    @defer.inlineCallbacks
    def test_messageReceived(self):
        self.svc = maildir.MaildirService(self.maildir)

        # add a fake messageReceived method
        messagesReceived = []

        def messageReceived(filename):
            messagesReceived.append(filename)
            return defer.succeed(None)
        self.svc.messageReceived = messageReceived
        yield self.svc.startService()

        self.assertEqual(messagesReceived, [])

        tmpfile = os.path.join(self.tmpdir, "newmsg")
        newfile = os.path.join(self.newdir, "newmsg")
        with open(tmpfile, "w", encoding='utf-8'):
            pass
        os.rename(tmpfile, newfile)

        # TODO: can we wait for a dnotify somehow, if enabled?
        yield self.svc.poll()

        self.assertEqual(messagesReceived, ['newmsg'])

    def test_moveToCurDir(self):
        self.svc = maildir.MaildirService(self.maildir)
        tmpfile = os.path.join(self.tmpdir, "newmsg")
        newfile = os.path.join(self.newdir, "newmsg")
        with open(tmpfile, "w", encoding='utf-8'):
            pass
        os.rename(tmpfile, newfile)
        f = self.svc.moveToCurDir("newmsg")
        f.close()
        self.assertEqual([os.path.exists(os.path.join(d, "newmsg"))
                          for d in (self.newdir, self.curdir, self.tmpdir)],
                         [False, True, False])



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
"""
Test clean shutdown functionality of the master
"""

import mock

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.worker.protocols.manager.base import BaseDispatcher
from buildbot.worker.protocols.manager.base import BaseManager


class FakeMaster:
    initLock = defer.DeferredLock()

    def addService(self, svc):
        pass

    @property
    def master(self):
        return self


class TestDispatcher(BaseDispatcher):
    def __init__(self, config_portstr, portstr):
        super().__init__(portstr)
        self.start_listening_count = 0
        self.stop_listening_count = 0

    def start_listening_port(self):
        def stopListening():
            self.stop_listening_count += 1

        self.start_listening_count += 1
        port = mock.Mock()
        port.stopListening = stopListening
        return port


class TestPort():
    def __init__(self, test):
        self.test = test

    def stopListening(self):
        self.test.stop_listening_count += 1


class TestManagerClass(BaseManager):
    dispatcher_class = TestDispatcher


class TestBaseManager(unittest.TestCase):

    @defer.inlineCallbacks
    def setUp(self):
        self.manager = TestManagerClass('test_base_manager')
        yield self.manager.setServiceParent(FakeMaster())

    def assert_equal_registration(self, result, expected):
        result_users = {key: result[key].users for key in result}
        self.assertEqual(result_users, expected)

    def assert_start_stop_listening_counts(self, disp, start_count, stop_count):
        self.assertEqual(disp.start_listening_count, start_count)
        self.assertEqual(disp.stop_listening_count, stop_count)

    @defer.inlineCallbacks
    def test_repr(self):
        reg = yield self.manager.register('tcp:port', 'x', 'y', 'pf')
        self.assertEqual(repr(self.manager.dispatchers['tcp:port']),
                         '<base.BaseDispatcher for x on tcp:port>')
        self.assertEqual(repr(reg), '<base.Registration for x on tcp:port>')

    @defer.inlineCallbacks
    def test_register_before_start_service(self):
        yield self.manager.register('tcp:port', 'user', 'pass', 'pf')

        self.assert_equal_registration(self.manager.dispatchers,
                                       {'tcp:port': {'user': ('pass', 'pf')}})

        disp = self.manager.dispatchers['tcp:port']
        self.assert_start_stop_listening_counts(disp, 0, 0)
        self.assertEqual(len(self.manager.services), 1)

        yield self.manager.startService()
        self.assert_start_stop_listening_counts(disp, 1, 0)

        yield self.manager.stopService()
        self.assert_start_stop_listening_counts(disp, 1, 1)

    @defer.inlineCallbacks
    def test_same_registration_two_times(self):
        yield self.manager.startService()
        yield self.manager.register('tcp:port', 'user', 'pass', 'pf')

        # one registration is ok
        self.assert_equal_registration(self.manager.dispatchers,
                                       {'tcp:port': {'user': ('pass', 'pf')}})
        self.assertEqual(len(self.manager.services), 1)

        disp = self.manager.dispatchers['tcp:port']
        self.assert_start_stop_listening_counts(disp, 1, 0)

        # same user is not allowed to register
        with self.assertRaises(KeyError):
            yield self.manager.register('tcp:port', 'user', 'pass', 'pf')

        yield self.manager.stopService()
        self.assert_start_stop_listening_counts(disp, 1, 1)

    @defer.inlineCallbacks
    def test_register_unregister_register(self):
        yield self.manager.startService()
        reg = yield self.manager.register('tcp:port', 'user', 'pass', 'pf')
        self.assert_equal_registration(self.manager.dispatchers,
                                       {'tcp:port': {'user': ('pass', 'pf')}})

        disp = self.manager.dispatchers['tcp:port']
        self.assert_start_stop_listening_counts(disp, 1, 0)

        reg.unregister()
        self.assert_equal_registration(self.manager.dispatchers, {})

        # allow registering same user again
        yield self.manager.register('tcp:port', 'user', 'pass', 'pf')

        self.assert_equal_registration(self.manager.dispatchers,
                                       {'tcp:port': {'user': ('pass', 'pf')}})

        yield self.manager.stopService()
        self.assert_start_stop_listening_counts(disp, 1, 1)

    @defer.inlineCallbacks
    def test_register_unregister_empty_disp_users(self):
        yield self.manager.startService()
        reg = yield self.manager.register('tcp:port', 'user', 'pass', 'pf')
        self.assertEqual(len(self.manager.services), 1)

        expected = {'tcp:port': {'user': ('pass', 'pf')}}
        self.assert_equal_registration(self.manager.dispatchers, expected)

        disp = self.manager.dispatchers['tcp:port']

        reg.unregister()
        self.assert_equal_registration(self.manager.dispatchers, {})
        self.assertEqual(reg.username, None)
        self.assertEqual(len(self.manager.services), 0)

        yield self.manager.stopService()
        self.assert_start_stop_listening_counts(disp, 1, 1)

    @defer.inlineCallbacks
    def test_different_ports_same_users(self):
        yield self.manager.startService()
        # same registrations on different ports is ok
        reg1 = yield self.manager.register('tcp:port1', "user", "pass", 'pf')
        reg2 = yield self.manager.register('tcp:port2', "user", "pass", 'pf')
        reg3 = yield self.manager.register('tcp:port3', "user", "pass", 'pf')

        disp1 = self.manager.dispatchers['tcp:port1']
        self.assert_start_stop_listening_counts(disp1, 1, 0)

        disp2 = self.manager.dispatchers['tcp:port2']
        self.assert_start_stop_listening_counts(disp2, 1, 0)

        disp3 = self.manager.dispatchers['tcp:port3']
        self.assert_start_stop_listening_counts(disp3, 1, 0)

        self.assert_equal_registration(self.manager.dispatchers, {
            'tcp:port1': {'user': ('pass', 'pf')},
            'tcp:port2': {'user': ('pass', 'pf')},
            'tcp:port3': {'user': ('pass', 'pf')}
        })
        self.assertEqual(len(self.manager.services), 3)

        yield reg1.unregister()
        self.assert_equal_registration(self.manager.dispatchers, {
            'tcp:port2': {'user': ('pass', 'pf')},
            'tcp:port3': {'user': ('pass', 'pf')}
        })
        self.assertEqual(reg1.username, None)
        self.assertEqual(len(self.manager.services), 2)
        self.assert_start_stop_listening_counts(disp1, 1, 1)

        yield reg2.unregister()
        self.assert_equal_registration(self.manager.dispatchers, {
            'tcp:port3': {'user': ('pass', 'pf')}
        })
        self.assertEqual(reg2.username, None)
        self.assertEqual(len(self.manager.services), 1)
        self.assert_start_stop_listening_counts(disp2, 1, 1)

        yield reg3.unregister()
        expected = {}
        self.assert_equal_registration(self.manager.dispatchers, expected)
        self.assertEqual(reg3.username, None)
        self.assertEqual(len(self.manager.services), 0)
        self.assert_start_stop_listening_counts(disp3, 1, 1)

        yield self.manager.stopService()
        self.assert_start_stop_listening_counts(disp1, 1, 1)
        self.assert_start_stop_listening_counts(disp2, 1, 1)
        self.assert_start_stop_listening_counts(disp3, 1, 1)

    @defer.inlineCallbacks
    def test_same_port_different_users(self):
        yield self.manager.startService()
        reg1 = yield self.manager.register('tcp:port', 'user1', 'pass1', 'pf1')
        reg2 = yield self.manager.register('tcp:port', 'user2', 'pass2', 'pf2')
        reg3 = yield self.manager.register('tcp:port', 'user3', 'pass3', 'pf3')
        disp = self.manager.dispatchers['tcp:port']

        self.assertEqual(len(self.manager.services), 1)
        self.assert_equal_registration(self.manager.dispatchers, {
            'tcp:port': {
                'user1': ('pass1', 'pf1'),
                'user2': ('pass2', 'pf2'),
                'user3': ('pass3', 'pf3')
            }
        })
        self.assertEqual(len(self.manager.services), 1)

        yield reg1.unregister()
        self.assert_equal_registration(self.manager.dispatchers, {
            'tcp:port': {
                'user2': ('pass2', 'pf2'),
                'user3': ('pass3', 'pf3')
            }
        })
        self.assertEqual(reg1.username, None)
        self.assertEqual(len(self.manager.services), 1)

        yield reg2.unregister()
        self.assert_equal_registration(self.manager.dispatchers, {
            'tcp:port': {
                'user3': ('pass3', 'pf3')
            }
        })
        self.assertEqual(reg2.username, None)
        self.assertEqual(len(self.manager.services), 1)

        yield reg3.unregister()
        expected = {}
        self.assert_equal_registration(self.manager.dispatchers, expected)
        self.assertEqual(reg3.username, None)
        self.assertEqual(len(self.manager.services), 0)

        yield self.manager.stopService()
        self.assert_start_stop_listening_counts(disp, 1, 1)
